# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the SSOWizardPage and related."""

from ubuntu_kylin_sso.qt import PREFERED_UI_SIZE, sso_wizard_page as gui
from ubuntu_kylin_sso.qt.tests import (
    APP_NAME,
    BaseTestCase,
    build_string_for_pixels,
    PageBaseTestCase,
)


MAX_WIDTH = 100


class WizardHeaderTestCase(BaseTestCase):

    """Tests for injected Header in each Wizard Page."""

    kwargs = dict(max_width=MAX_WIDTH)
    ui_class = gui.WizardHeader
    ui_wizard_class = None

    def test_label_state(self):
        """Check the title and subtitle properties."""
        self.assertTrue(self.ui.title_label.wordWrap())
        self.assertTrue(self.ui.subtitle_label.wordWrap())
        self.assertFalse(self.ui.title_label.isVisible())
        self.assertFalse(self.ui.subtitle_label.isVisible())

    def test_set_title(self):
        """Check if set_title works ok, showing the widget if necessary."""
        max_width = self.ui.max_title_width
        text = build_string_for_pixels(self.ui.title_label, max_width)

        self.ui.set_title(text)

        self.assert_title_correct(self.ui.title_label, text, max_width)

    def test_set_elided_title(self):
        """Check if set_title adds the ellipsis when necessary."""
        # add an extra letter so we ensure this needs to be trimmed
        max_width = self.ui.max_title_width
        text = build_string_for_pixels(self.ui.title_label, max_width + 10)

        self.ui.set_title(text)

        self.assert_title_correct(self.ui.title_label, text, max_width)

    def test_set_empty_title(self):
        """Check if the widget is hidden for empty title."""
        self.ui.set_title('')

        self.assertEqual(self.ui.title_label.toolTip(), '')
        self.assertFalse(self.ui.title_label.isVisible())

    def test_set_subtitle(self):
        """Check if set_subtitle works ok, showing the widget if necessary."""
        max_width = self.ui.max_subtitle_width
        text = build_string_for_pixels(self.ui.subtitle_label, max_width)

        self.ui.set_subtitle(text)

        self.assert_subtitle_correct(self.ui.subtitle_label, text, max_width)

    def test_set_elided_subtitle(self):
        """Check if set_subtitle adds the ellipsis when necessary."""
        max_width = self.ui.max_subtitle_width
        text = build_string_for_pixels(self.ui.subtitle_label, max_width + 10)

        self.ui.set_subtitle(text)

        self.assert_subtitle_correct(self.ui.subtitle_label, text, max_width)

    def test_set_empty_subtitle(self):
        """Check if the widget is hidden for empty subtitle."""
        self.ui.set_subtitle('')

        self.assertEqual(self.ui.subtitle_label.toolTip(), '')
        self.assertFalse(self.ui.subtitle_label.isVisible())


class BaseWizardPageTestCase(PageBaseTestCase):

    """Tests for SSOWizardPage."""

    kwargs = {}
    ui_class = gui.BaseWizardPage

    def test_max_width(self):
        """The max_width is correct."""
        self.assertEqual(self.ui.max_width, 0)


class SSOWizardPageTestCase(BaseWizardPageTestCase):

    """Tests for SSOWizardPage."""

    kwargs = dict(app_name=APP_NAME)
    ui_class = gui.SSOWizardPage

    def test_max_width(self):
        """The max_width is correct."""
        self.assertEqual(self.ui.max_width, PREFERED_UI_SIZE['width'])

    def test_show_error(self):
        """Test show_error with a normal length string."""
        message = 'error-message'
        self.ui.show_error(message)

        self.assert_error_correct(self.ui.form_errors_label, message,
                                  self.ui.header.max_title_width)

    def test_show_error_long_text(self):
        """Test show_error with a long length string."""
        message = build_string_for_pixels(self.ui.form_errors_label,
                                          self.ui.header.max_title_width + 10)

        self.ui.show_error(message)
        self.assert_error_correct(self.ui.form_errors_label, message,
                                  self.ui.header.max_title_width)

    def test_hide_error(self):
        """Test show_error with a long length string."""
        self.ui.hide_error()

        self.assertEqual(self.ui.form_errors_label.text(), ' ')

    def test_setup_page_with_failing_backend(self):
        """Test how the ui react with an invalid backend."""
        self.patch(gui.main, "get_sso_client", lambda: None)
        self.patch(self.ui, "show_error", self._set_called)
        self.ui.setup_page()
        reason = 'There was a problem trying to setup the page %r' % self.ui
        expected = ((reason,), {})
        self.assertEqual(expected, self._called)
        self.assertFalse(self.ui.isEnabled())



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Customized Setup Account page for SSO."""

import tempfile
import os
from io import BytesIO
from functools import partial

# pylint: disable=F0401
try:
    from PIL import Image
except ImportError:
    import Image
# pylint: enable=F0401

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso import NO_OP
from ubuntu_kylin_sso.logger import setup_gui_logging, log_call
from ubuntu_kylin_sso.qt import (
    LINK_STYLE,
    build_general_error_message,
    common,
    enhanced_check_box,
    ERROR_STYLE,
)
from ubuntu_kylin_sso.qt.sso_wizard_page import SSOWizardEnhancedEditPage
from ubuntu_kylin_sso.qt.ui.setup_account_ui import Ui_SetUpAccountPage
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    AGREE_TO_PRIVACY_POLICY,
    AGREE_TO_TERMS,
    AGREE_TO_TERMS_AND_PRIVACY_POLICY,
    CAPTCHA_LOAD_ERROR,
    CAPTCHA_RELOAD_MESSAGE,
    CAPTCHA_RELOAD_TEXT,
    CAPTCHA_REQUIRED_ERROR,
    CAPTCHA_SOLUTION_ENTRY,
    EMAIL,
    EMAIL1_ENTRY,
    EMAIL2_ENTRY,
    EMAIL_INVALID,
    EMAIL_MATCH,
    EMAIL_MISMATCH,
    EMPTY_NAME,
    INVALID_EMAIL,
    is_min_required_password,
    is_correct_email,
    JOIN_HEADER_LABEL,
    NAME_ENTRY,
    NAME_INVALID,
    PASSWORD,
    PASSWORD1_ENTRY,
    PASSWORD2_ENTRY,
    PASSWORD_HELP,
    PASSWORD_MISMATCH,
    PASSWORD_TOO_WEAK,
    PRIVACY_POLICY_TEXT,
    RETYPE_EMAIL,
    RETYPE_PASSWORD,
    SET_UP_ACCOUNT_BUTTON,
    SIGN_IN_LABEL,
    TERMS_TEXT,
    REGISTER_TITLE,
)


logger = setup_gui_logging('ubuntu_kylin_sso.setup_account_page')

ERROR_EMAIL = 'email'


class SetupAccountPage(SSOWizardEnhancedEditPage):
    """Customized Setup Account page for SSO."""

    ui_class = Ui_SetUpAccountPage
    userRegistered = QtCore.pyqtSignal(compat.text_type)
    signIn = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        self.captcha_file = None
        self.captcha_id = None
        self.captcha_received = False
        self.set_up_button = None
        self.terms_checkbox = None
        super(SetupAccountPage, self).__init__(*args, **kwargs)

    @property
    def _signals(self):
        """The signals to connect to the backend."""
        result = {
            'CaptchaGenerated':
                self._filter_by_app_name(self.on_captcha_generated),
            'CaptchaGenerationError':
                self._filter_by_app_name(self.on_captcha_generation_error),
            'UserRegistered':
                self._filter_by_app_name(self.on_user_registered),
            'UserRegistrationError':
                self._filter_by_app_name(self.on_user_registration_error),
        }
        return result

    @property
    def password(self):
        """Return the content of the password edit."""
        return compat.text_type(self.ui.password_edit.text())

    # Invalid name "initializePage"
    # pylint: disable=C0103

    def initializePage(self):
        """Setup UI details."""
        logger.debug('initializePage - About to show SetupAccountPage')
        # Set Setup Account button
        self.wizard().setOption(QtGui.QWizard.HaveCustomButton3, True)
        try:
            self.wizard().customButtonClicked.disconnect()
        except TypeError:
            pass
        self.setButtonText(QtGui.QWizard.CustomButton3, SET_UP_ACCOUNT_BUTTON)
        self.set_up_button = self.wizard().button(QtGui.QWizard.CustomButton3)
        self.set_up_button.clicked.connect(self.set_next_validation)
        self.set_up_button.setEnabled(False)

        self.ui.name_label.setText(NAME_ENTRY)
        self.ui.email_label.setText(EMAIL)
        self.ui.confirm_email_label.setText(RETYPE_EMAIL)
        self.ui.password_label.setText(PASSWORD)
        self.ui.confirm_password_label.setText(RETYPE_PASSWORD)

        # Button setup
        self.wizard().setButtonLayout([
            QtGui.QWizard.Stretch,
            QtGui.QWizard.CustomButton3])

        common.password_default_assistance(self.ui.password_assistance)
        # Hide assistance labels by default
        self.ui.name_assistance.setVisible(False)
        self.ui.email_assistance.setVisible(False)
        self.ui.confirm_email_assistance.setVisible(False)
        self.ui.password_assistance.setVisible(False)
        self.ui.refresh_label.setVisible(True)

    # pylint: enable=C0103

    def _set_translated_strings(self):
        """Set the strings."""
        # set the translated string
        title_page = REGISTER_TITLE.format(app_name=self.app_name)
        self.setTitle(title_page)
        self.setSubTitle(self.help_text)

        self.ui.name_label.setText(NAME_ENTRY)
        sign_in_link = LINK_STYLE.format(link_url='#',
            link_text=SIGN_IN_LABEL)
        self.ui.sign_in_label.setText(sign_in_link)
        self.ui.email_label.setText(EMAIL1_ENTRY)
        self.ui.confirm_email_label.setText(EMAIL2_ENTRY)
        self.ui.password_label.setText(PASSWORD1_ENTRY)
        self.ui.confirm_password_label.setText(PASSWORD2_ENTRY)
        self.ui.password_edit.setToolTip(PASSWORD_HELP)
        self.ui.captcha_solution_edit.setPlaceholderText(
            CAPTCHA_SOLUTION_ENTRY)
        link = LINK_STYLE.format(link_url='#', link_text=CAPTCHA_RELOAD_TEXT)
        self.ui.refresh_label.setText(CAPTCHA_RELOAD_MESSAGE %
                                      {'reload_link': link})

        if self.tc_url:
            terms_links = LINK_STYLE.format(link_url=self.tc_url,
                                            link_text=TERMS_TEXT)
        if self.policy_url:
            privacy_policy_link = LINK_STYLE.format(link_url=self.policy_url,
                                                 link_text=PRIVACY_POLICY_TEXT)

        terms = ''
        if self.tc_url and self.policy_url:
            terms = AGREE_TO_TERMS_AND_PRIVACY_POLICY.format(
                        app_name=self.app_name,
                        terms_and_conditions=terms_links,
                        privacy_policy=privacy_policy_link)
        elif self.tc_url:
            terms = AGREE_TO_TERMS.format(app_name=self.app_name,
                        terms_and_conditions=terms_links)
        elif self.policy_url:
            terms = AGREE_TO_PRIVACY_POLICY.format(app_name=self.app_name,
                        privacy_policy=privacy_policy_link)

        self.terms_checkbox = enhanced_check_box.EnhancedCheckBox(terms, self)
        self.ui.hlayout_check.addWidget(self.terms_checkbox)
        self.terms_checkbox.setVisible(bool(self.tc_url or self.policy_url))

        self._register_fields()

    def _set_line_edits_validations(self):
        """Set the validations to be performed on the edits."""
        logger.debug('SetUpAccountPage._set_line_edits_validations')
        self.set_line_edit_validation_rule(self.ui.email_edit,
                                                is_correct_email)
        # set the validation rule for the email confirmation
        self.set_line_edit_validation_rule(
                                        self.ui.confirm_email_edit,
                                        self.is_correct_email_confirmation)
        # connect the changed text of the password to trigger a changed text
        # in the confirm so that the validation is redone
        self.ui.email_edit.textChanged.connect(
                            self.ui.confirm_email_edit.textChanged.emit)
        self.set_line_edit_validation_rule(self.ui.password_edit,
                                                is_min_required_password)
        self.set_line_edit_validation_rule(
                                        self.ui.confirm_password_edit,
                                        self.is_correct_password_confirmation)
        # same as the above case, lets connect a signal to a signal
        self.ui.password_edit.textChanged.connect(
            self.ui.confirm_password_edit.textChanged.emit)

    def _connect_ui(self):
        """Set the connection of signals."""
        self._set_line_edits_validations()

        self.ui.captcha_view.setPixmap(QtGui.QPixmap())
        self.ui.password_edit.textEdited.connect(
            lambda: common.password_assistance(self.ui.password_edit,
                                                 self.ui.password_assistance,
                                                 common.NORMAL))

        self.ui.refresh_label.linkActivated.connect(self.hide_error)
        self.ui.refresh_label.linkActivated.connect(
            lambda url: self._refresh_captcha())
        # We need to check if we enable the button on many signals
        self.ui.name_edit.textEdited.connect(self._enable_setup_button)
        self.ui.email_edit.textEdited.connect(self._enable_setup_button)
        self.ui.confirm_email_edit.textEdited.connect(
            self._enable_setup_button)
        self.ui.password_edit.textEdited.connect(self._enable_setup_button)
        self.ui.confirm_password_edit.textEdited.connect(
            self._enable_setup_button)
        self.ui.captcha_solution_edit.textEdited.connect(
            self._enable_setup_button)
        self.terms_checkbox.stateChanged.connect(self._enable_setup_button)
        self.ui.sign_in_label.linkActivated.connect(self.signIn)

        self._refresh_captcha()

    def _enable_setup_button(self):
        """Only enable the setup button if the form is valid."""
        name = compat.text_type(self.ui.name_edit.text()).strip()
        email = compat.text_type(self.ui.email_edit.text())
        confirm_email = compat.text_type(self.ui.confirm_email_edit.text())
        password = compat.text_type(self.ui.password_edit.text())
        confirm_password = compat.text_type(
                self.ui.confirm_password_edit.text())
        captcha_solution = compat.text_type(
                self.ui.captcha_solution_edit.text())

        # Check for len(name) > 0 to ensure that a bool is assigned to enabled
        if not self.terms_checkbox.isVisible():
            checkbox_terms = True
        else:
            checkbox_terms = self.terms_checkbox.isChecked()

        enabled = checkbox_terms and \
          len(captcha_solution) > 0 and \
          is_min_required_password(password) and \
          password == confirm_password and is_correct_email(email) and \
          email == confirm_email and len(name) > 0

        self.set_up_button.setEnabled(enabled)

    def _refresh_captcha(self):
        """Refresh the captcha image shown in the ui."""
        logger.debug('SetUpAccountPage._refresh_captcha')
        # lets clean behind us, do we have the old file arround?
        if self.captcha_file and os.path.exists(self.captcha_file):
            os.unlink(self.captcha_file)
        fd = tempfile.NamedTemporaryFile()
        file_name = fd.name
        self.captcha_file = file_name
        args = (self.app_name, file_name)
        f = self.backend.generate_captcha
        error_handler = partial(self._handle_error, f,
            self.on_captcha_generation_error)
        f(*args, reply_handler=NO_OP, error_handler=error_handler)
        self.on_captcha_refreshing()

    def _set_titles(self):
        """Set the diff titles of the view."""
        logger.debug('SetUpAccountPage._set_titles')
        self.header.set_title(
            JOIN_HEADER_LABEL % {'app_name': self.app_name})
        self.header.set_subtitle(self._subtitle)

    def _register_fields(self):
        """Register the diff fields of the Ui."""
        self.registerField('email_address', self.ui.email_edit)
        self.registerField('password', self.ui.password_edit)

    @log_call(logger.debug)
    def on_captcha_generated(self, app_name, result):
        """A new image was generated."""
        self.captcha_id = result
        # HACK: First, let me apologize before hand, you can mention my mother
        # if needed I would do the same (mandel)
        # In an ideal world we could use the Qt plug-in for the images so that
        # we could load jpgs etc.. but this won't work when the app has been
        # brozen win py2exe using bundle_files=1
        # The main issue is that Qt will complain about the thread not being
        # the correct one when performing a moveToThread operation which is
        # done either by a setParent or something within the qtreactor, PIL
        # in this case does solve the issue. Sorry :(
        pil_image = Image.open(self.captcha_file)
        bytes_io = BytesIO()
        pil_image.save(bytes_io, format='png')
        pixmap_image = QtGui.QPixmap()
        pixmap_image.loadFromData(bytes_io.getvalue())
        self.captcha_image = pixmap_image
        self.on_captcha_refresh_complete()

    @log_call(logger.error)
    def on_captcha_generation_error(self, app_name, error):
        """An error ocurred."""
        self.show_error(CAPTCHA_LOAD_ERROR)
        self.on_captcha_refresh_complete()

    @log_call(logger.error)
    def on_user_registration_error(self, app_name, error):
        """Let the user know we could not register."""
        # errors are returned as a dict with the data we want to show.
        msg = error.pop(ERROR_EMAIL, '')
        if msg:
            self.set_error_message(self.ui.email_assistance, msg)
        error_msg = build_general_error_message(error)
        if error_msg:
            self.show_error(error_msg)
        self._refresh_captcha()

    @log_call(logger.info)
    def on_user_registered(self, app_name, email):
        """Execute when the user did register."""
        self.hide_overlay()
        email = compat.text_type(self.ui.email_edit.text())
        self.userRegistered.emit(email)

    def validate_form(self):
        """Validate the info of the form and return an error."""
        logger.debug('SetUpAccountPage.validate_form')
        name = compat.text_type(self.ui.name_edit.text()).strip()
        email = compat.text_type(self.ui.email_edit.text())
        confirm_email = compat.text_type(self.ui.confirm_email_edit.text())
        password = compat.text_type(self.ui.password_edit.text())
        confirm_password = compat.text_type(
                self.ui.confirm_password_edit.text())
        captcha_solution = compat.text_type(
                self.ui.captcha_solution_edit.text())
        condition = True
        messages = []
        if not name:
            condition = False
            self.set_error_message(self.ui.name_assistance,
                NAME_INVALID)
        if not is_correct_email(email):
            condition = False
            self.set_error_message(self.ui.email_assistance,
                EMAIL_INVALID)
        if email != confirm_email:
            condition = False
            self.set_error_message(self.ui.confirm_email_assistance,
                EMAIL_MISMATCH)
        if not is_min_required_password(password):
            messages.append(PASSWORD_TOO_WEAK)
        if password != confirm_password:
            messages.append(PASSWORD_MISMATCH)
        if not captcha_solution:
            messages.append(CAPTCHA_REQUIRED_ERROR)
        if len(messages) > 0:
            condition = False
            self.show_error('\n'.join(messages))
        return condition

    def set_next_validation(self):
        """Set the validation as the next page."""
        logger.debug('SetUpAccountPage.set_next_validation')
        email = compat.text_type(self.ui.email_edit.text())
        password = compat.text_type(self.ui.password_edit.text())
        name = compat.text_type(self.ui.name_edit.text())
        captcha_id = self.captcha_id
        captcha_solution = compat.text_type(
                self.ui.captcha_solution_edit.text())
        # validate the current info of the form, try to perform the action
        # to register the user, and then move foward
        if self.validate_form():
            self.show_overlay()
            self.hide_error()
            args = (self.app_name, email, password, name, captcha_id,
                captcha_solution)
            f = self.backend.register_user
            error_handler = partial(self._handle_error, f,
                self.on_user_registration_error)
            f(*args, reply_handler=NO_OP, error_handler=error_handler)

    def is_correct_email(self, email_address):
        """Return if the email is correct."""
        return '@' in email_address

    def is_correct_email_confirmation(self, email_address):
        """Return that the email is the same."""
        return compat.text_type(self.ui.email_edit.text()) == email_address

    def is_correct_password_confirmation(self, password):
        """Return that the passwords are correct."""
        return compat.text_type(self.ui.password_edit.text()) == password

    def focus_changed(self, old, now):
        """Check who has the focus to activate password popups if necessary."""
        if old == self.ui.name_edit:
            self.name_assistance()
        elif old == self.ui.email_edit:
            self.email_assistance()
        elif old == self.ui.confirm_email_edit:
            self.confirm_email_assistance()
        elif old == self.ui.confirm_password_edit:
            common.password_check_match(self.ui.password_edit,
                                      self.ui.confirm_password_edit,
                                      self.ui.password_assistance)
        if now == self.ui.password_edit:
            self.ui.password_assistance.setVisible(True)

    def name_assistance(self):
        """Show help for the name field."""
        text = compat.text_type(self.ui.name_edit.text())
        if not text.strip():
            self.set_error_message(self.ui.name_assistance,
                EMPTY_NAME)
            common.check_as_invalid(self.ui.name_edit)
        else:
            self.ui.name_assistance.setVisible(False)
            common.check_as_valid(self.ui.name_edit)

    def email_assistance(self):
        """Show help for the email field."""
        text = compat.text_type(self.ui.email_edit.text())
        if not is_correct_email(text):
            self.set_error_message(self.ui.email_assistance,
                INVALID_EMAIL)
            common.check_as_invalid(self.ui.email_edit)
        else:
            self.ui.email_assistance.setVisible(False)
            common.check_as_valid(self.ui.email_edit)

    def confirm_email_assistance(self):
        """Show help for the confirm email field."""
        text1 = compat.text_type(self.ui.email_edit.text())
        text2 = compat.text_type(self.ui.confirm_email_edit.text())
        if text1 != text2:
            self.set_error_message(self.ui.confirm_email_assistance,
                EMAIL_MATCH)
            common.check_as_invalid(self.ui.confirm_email_edit)
        else:
            self.ui.confirm_email_assistance.setVisible(False)
            common.check_as_valid(self.ui.confirm_email_edit)

    def set_error_message(self, label, msg):
        """Set the message to the proper label applying the error style."""
        label.setText(ERROR_STYLE % msg)
        label.setVisible(True)

    # pylint: disable=C0103

    def showEvent(self, event):
        """Set set_up_button as default button when the page is shown."""
        # This method should stays here because if we move it to initializePage
        # set_up_button won't take the proper style for hover and press
        if self.set_up_button is not None:
            self.set_up_button.setVisible(True)
            self.set_up_button.setDefault(True)
        self.connect(QtGui.QApplication.instance(),
            QtCore.SIGNAL("focusChanged(QWidget*, QWidget*)"),
            self.focus_changed)
        super(SetupAccountPage, self).showEvent(event)
        if not self.captcha_received:
            self.show_overlay()

    def hideEvent(self, event):
        """Disconnect the focusChanged signal when the page change."""
        if self.set_up_button is not None:
            self.set_up_button.setVisible(False)
        try:
            self.disconnect(QtGui.QApplication.instance(),
                QtCore.SIGNAL("focusChanged(QWidget*, QWidget*)"),
                self.focus_changed)
        except TypeError:
            pass
        super(SetupAccountPage, self).hideEvent(event)

    # pylint: enable=C0103

    def on_captcha_refreshing(self):
        """Show overlay when captcha is refreshing."""
        logger.info('SetUpAccountPage.on_captcha_refreshing')
        if self.isVisible():
            self.show_overlay()
        self.captcha_received = False

    def on_captcha_refresh_complete(self):
        """Hide overlay when captcha finished refreshing."""
        logger.info('SetUpAccountPage.on_captcha_refresh_complete')
        self.hide_overlay()
        self.captcha_received = True

    def get_captcha_image(self):
        """Return the path to the captcha image."""
        return self.ui.captcha_view.pixmap()

    def set_captcha_image(self, pixmap_image):
        """Set the new image of the captcha."""
        # lets set the QPixmap for the label
        self.ui.captcha_view.setPixmap(pixmap_image)

    captcha_image = property(get_captcha_image, set_captcha_image)



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Keyring implementation for Windows and Darwin."""

from json import loads, dumps

from ubuntu_kylin_sso.utils.qtwisted import qtDeferToThread as deferToThread

USERNAME = 'ubuntu_kylin_sso'


class Keyring(object):
    """A Keyring for a given application name."""

    def __init__(self, keyring=None):
        """Create a new instance."""
        if keyring is None:
            import keyring as pykeyring
            keyring = pykeyring
        self.keyring = keyring

    def set_credentials(self, app_name, cred):
        """Set the credentials of the Ubuntu SSO item."""
        # the windows keyring can only store a pair username-password
        # so we store the data using ubuntu_kylin_sso as the user name. Then
        # the cred will be stored as the string representation of the dict.
        return deferToThread(self.keyring.set_password, app_name, USERNAME,
                             dumps(cred))

    def _get_credentials_obj(self, app_name):
        """A dict with the credentials."""
        creds = self.keyring.get_password(app_name, USERNAME)
        if creds:
            return loads(creds)

    def get_credentials(self, app_name):
        """A deferred with the secret of the SSO item in a dictionary."""
        return deferToThread(self._get_credentials_obj, app_name)

    def delete_credentials(self, app_name):
        """Delete a set of credentials from the keyring."""
        # this call depends on a patch I sent to pykeyring. The patch has
        # not landed as of version 0.5.1. If you have that version you can
        # clone my patch in the following way:
        # hg clone https://bitbucket.org/mandel/pykeyring-delete-password
        # pylint: disable=E1103
        return deferToThread(self.keyring.delete_password, app_name, USERNAME)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the Reset Password Page."""

# pylint: disable=F0401,E0611,E1101

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    PASSWORD1_ENTRY,
    PASSWORD2_ENTRY,
    RESET_CODE_ENTRY,
)
from ubuntu_kylin_sso.qt import common
from ubuntu_kylin_sso.qt.reset_password_page import (
    ResetPasswordPage,
    RESET_SUBTITLE,
    RESET_TITLE,
)
from ubuntu_kylin_sso.qt.tests import PageBaseTestCase


# We need this Fake until a future refactor.
class FakeForgottenPage(object):

    """Fake Forgotten Page."""

    def __init__(self):
        self.ui = self
        self.email_line_edit = self
        self.line_text = ''

    # pylint: disable=C0103
    def setText(self, text):
        """Fake setText for Line edit."""
        self.line_text = text
    # pylint: enable=C0103

    def text(self):
        """Fake text for line edit."""
        return self.line_text


class ResetPasswordTestCase(PageBaseTestCase):
    """Test the ResetPasswordPage code."""

    ui_class = ResetPasswordPage
    ui_signals = ('passwordChanged',)
    ui_backend_siganls = ('PasswordChanged', 'PasswordChangeError')

    def test_password_line(self):
        """Check the initial state of ResetPassword."""
        self.assertEqual(self.ui.ui.password_line_edit.receivers(
            QtCore.SIGNAL('textEdited(QString)')), 1)
        self.assertEqual(self.ui.ui.confirm_password_line_edit.receivers(
            QtCore.SIGNAL('textEdited(QString)')), 1)

    def test_initialize(self):
        """Check the Title and Subtitle."""
        self.ui.initializePage()
        self.assert_title_correct(RESET_TITLE)
        self.assert_subtitle_correct(RESET_SUBTITLE)
        self.assertEqual(compat.text_type(self.ui.ui.password_label.text()),
            PASSWORD1_ENTRY)
        self.assertEqual(
                compat.text_type(self.ui.ui.confirm_password_label.text()),
                PASSWORD2_ENTRY)
        self.assertEqual(compat.text_type(self.ui.ui.reset_code.text()),
            RESET_CODE_ENTRY)

    def test_focus_changed_password_visibility(self):
        """Check visibility changes when focus_changed() is executed."""
        self.ui.focus_changed(None, self.ui.ui.password_line_edit)
        self.assertTrue(self.ui.ui.password_assistance.isVisible())

    def test_show_hide_event(self):
        """Check connections to focusChanged on show and hide event."""
        self.assertEqual(QtGui.QApplication.instance().receivers(
            QtCore.SIGNAL("focusChanged(QWidget*, QWidget*)")), 1)
        self.ui.hide()
        self.assertEqual(QtGui.QApplication.instance().receivers(
            QtCore.SIGNAL("focusChanged(QWidget*, QWidget*)")), 0)

    # pylint: disable=W0212

    def test_focus_changed_1(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(common, 'password_default_assistance', self._set_called)
        self.assertFalse(self._called)

        self.ui.focus_changed(None, self.ui.ui.password_line_edit)

        self.assertTrue(self.ui.ui.password_assistance.isVisible())
        self.assertTrue(self._called)

    def test_focus_changed_2(self):
        """Check functions execution when focus_changed() is executed."""
        self.patch(common, 'password_check_match', self._set_called)
        self.assertFalse(self._called)

        self.ui.focus_changed(None, self.ui.ui.confirm_password_line_edit)

        self.assertTrue(self.ui.ui.password_assistance.isVisible())
        self.assertTrue(self._called)

    # pylint: enable=W0212

    def test_set_new_password(self):
        """Test set_new_password method."""
        self.patch(self.ui, "hide_error", self._set_called)
        email = 'email@example.com'
        code = 'code'
        password = 'password'
        forgotten = FakeForgottenPage()
        forgotten.setText(email)
        self.patch(self.wizard, "forgotten", forgotten)
        self.ui.ui.reset_code_line_edit.setText(code)
        self.ui.ui.password_line_edit.setText(password)

        self.ui.set_new_password()

        self.assert_backend_called('set_new_password',
            self.app_name, email, code, password)
        expected = ((), {})
        self.assertEqual(expected, self._called)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the runner helper module."""

from __future__ import unicode_literals

import os

from twisted.internet import defer

from ubuntu_kylin_sso.tests import TestCase
from ubuntu_kylin_sso.utils import runner


TEST_ME_DIR = 'test-me-more-♥'
TEST_ME_DIR_BYTES = TEST_ME_DIR.encode('utf-8')


class SpawnProgramTestCase(TestCase):
    """The test suite for the spawn_program method."""

    args = ('python', '-c',
            'import os; os.system("mkdir %s")' % TEST_ME_DIR)
    use_reactor = True
    timeout = 3

    @defer.inlineCallbacks
    def setUp(self):
        yield super(SpawnProgramTestCase, self).setUp()
        assert not os.path.exists(TEST_ME_DIR_BYTES)
        self.addCleanup(lambda: os.path.exists(TEST_ME_DIR_BYTES) and
                                os.rmdir(TEST_ME_DIR_BYTES))

    def spawn_fn(self, args):
        """The target function to test."""
        return runner.spawn_program(args, use_reactor=self.use_reactor)

    def assert_command_was_run(self):
        """The spawnned commnad was correctly run."""
        self.assertTrue(os.path.exists(TEST_ME_DIR_BYTES))
        self.assertTrue(os.path.isdir(TEST_ME_DIR_BYTES))

    @defer.inlineCallbacks
    def test_program_is_spawned(self):
        """The program is actually spawned."""
        yield self.spawn_fn(self.args)
        self.assert_command_was_run()

    @defer.inlineCallbacks
    def test_program_is_spawned_returned_code_non_zero(self):
        """The program is actually spawned."""
        status = yield self.spawn_fn(self.args)
        self.assertEqual(status, 0)

    @defer.inlineCallbacks
    def test_failed_to_start(self):
        """FailedToStartError is raised if the program does not start."""
        no_such_program = './myexecutablethatdoesnotexist'
        assert not os.path.exists(no_such_program)

        d = self.spawn_fn((no_such_program,))
        exc = yield self.assertFailure(d, runner.FailedToStartError)

        self.assertIn(no_such_program, exc.message)

    @defer.inlineCallbacks
    def test_other_error(self):
        """SpawnError is raised if the program does not start."""
        d = self.spawn_fn((None,))
        exc = yield self.assertFailure(d, runner.SpawnError)

        self.assertIn('Unspecified error', exc.message)



# -*- coding: utf-8 -*-

# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""
Provides a twisted interface to access the system keyring via DBus.
Implements the Secrets Service API Draft:
 * http://code.confuego.org/secrets-xdg-specs/
"""

import dbus

from twisted.internet.defer import Deferred

BUS_NAME = "org.freedesktop.secrets"
SERVICE_IFACE = "org.freedesktop.Secret.Service"
PROMPT_IFACE = "org.freedesktop.Secret.Prompt"
SESSION_IFACE = "org.freedesktop.Secret.Session"
COLLECTION_IFACE = "org.freedesktop.Secret.Collection"
ITEM_IFACE = "org.freedesktop.Secret.Item"
PROPERTIES_IFACE = "org.freedesktop.DBus.Properties"
SECRETS_SERVICE = "/org/freedesktop/secrets"
DEFAULT_COLLECTION = "/org/freedesktop/secrets/aliases/default"
SESSION_COLLECTION = "/org/freedesktop/secrets/collection/session"
SECRET_CONTENT_TYPE = "application/octet-stream"

ALGORITHM = "plain"
ALGORITHM_PARAMS = ""
CLXN_LABEL_PROPERTY = "org.freedesktop.Secret.Collection.Label"
CLXN_LABEL_PROPERTY_OLD = "Label"
ITEM_LABEL_PROPERTY = "org.freedesktop.Secret.Item.Label"
ITEM_LABEL_PROPERTY_OLD = "Label"
ITEM_ATTRIBUTES_PROPERTY = "org.freedesktop.Secret.Item.Attributes"
ITEM_ATTRIBUTES_PROPERTY_OLD = "Attributes"
COLLECTIONS_PROPERTY = "org.freedesktop.Secret.Service.Collections"
COLLECTIONS_PROPERTY_OLD = "Collections"
DEFAULT_LABEL = "default"


class UserCancelled(Exception):
    """The user cancelled a prompt."""


def no_op(*args):
    """Do nothing."""


class SecretService(object):
    """The Secret Service manages all the sessions and collections."""
    service = None
    properties = None
    session = None
    bus = None
    window_id = None

    def open_session(self, window_id=0):
        """Open a unique session for the caller application."""
        d = Deferred()
        try:
            self.window_id = str(window_id)
            self.bus = dbus.SessionBus()
            service_object = self.bus.get_object(BUS_NAME, SECRETS_SERVICE)
            self.service = dbus.Interface(service_object,
                                          dbus_interface=SERVICE_IFACE)
            self.properties = dbus.Interface(service_object,
                                             dbus_interface=PROPERTIES_IFACE)

            def session_opened(result, session):
                """The session was successfully opened."""
                self.session = self.bus.get_object(BUS_NAME, session)
                d.callback(self)

            parameters = dbus.String(ALGORITHM_PARAMS, variant_level=1)
            self.service.OpenSession(ALGORITHM, parameters,
                                     reply_handler=session_opened,
                                     error_handler=d.errback)
        except dbus.exceptions.DBusException as e:
            d.errback(e)
        return d

    def do_prompt(self, prompt_path):
        """Show a prompt given its path."""
        d = Deferred()
        prompt_object = self.bus.get_object(BUS_NAME, prompt_path)
        prompt = dbus.Interface(prompt_object, dbus_interface=PROMPT_IFACE)

        def prompt_completed(dismissed, result):
            """The prompt was either completed or dismissed."""
            sigcompleted.remove()
            if dismissed:
                d.errback(UserCancelled())
            else:
                d.callback(result)

        sigcompleted = prompt.connect_to_signal("Completed", prompt_completed)
        prompt.Prompt(self.window_id,
                      reply_handler=no_op,
                      error_handler=d.errback)
        return d

    def make_item_list(self, object_path_list):
        """Make a list of items given their paths."""
        return [Item(self, o) for o in object_path_list]

    def search_items(self, attributes):
        """Find items in any collection."""
        d = Deferred()
        result = []

        def prompt_handle(unlocked):
            """Merge the items that were just unlocked."""
            result.extend(unlocked)
            return result

        def unlock_handler(unlocked, prompt):
            """The items were unlocked, or a prompt should be shown first."""
            result.extend(unlocked)
            if prompt != "/":
                d2 = self.do_prompt(prompt)
                d2.addCallback(prompt_handle)
                d2.chainDeferred(d)
            else:
                d.callback(result)

        def items_found(unlocked, locked):
            """Called with two lists of found items."""
            result.extend(unlocked)
            if len(locked) > 0:
                self.service.Unlock(locked,
                                    reply_handler=unlock_handler,
                                    error_handler=d.errback)
            else:
                d.callback(result)

        self.service.SearchItems(attributes,
                                 reply_handler=items_found,
                                 error_handler=d.errback)
        d.addCallback(self.make_item_list)
        return d

    def create_collection(self, label, alias=''):
        """Create a new collection with the specified properties."""
        d = Deferred()

        def createcollection_handler(collection, prompt):
            """A collection was created, or a prompt should be shown first."""
            if prompt != "/":
                self.do_prompt(prompt).chainDeferred(d)
            else:
                d.callback(collection)

        def error_fallback(error):
            """Fall back to using the old property name."""
            properties = {CLXN_LABEL_PROPERTY_OLD: dbus.String(
                    label,
                    variant_level=1)}
            self.service.CreateCollection(
                properties,
                reply_handler=createcollection_handler,
                error_handler=d.errback)

        properties = {CLXN_LABEL_PROPERTY: dbus.String(label,
                                                       variant_level=1)}
        try:
            self.service.CreateCollection(
                properties, alias,
                reply_handler=createcollection_handler,
                error_handler=error_fallback)
        except TypeError:
            error_fallback(None)

        d.addCallback(lambda p: Collection(self, p))
        return d

    def get_collections(self):
        """Return the list of all collections."""
        d = Deferred()

        def propertyget_handler(collection_paths):
            """The list of collection paths was retrieved."""
            result = []
            for path in collection_paths:
                collection = Collection(self, path)
                result.append(collection)
            d.callback(result)

        def error_fallback(error):
            """Fall back to the old property name."""
            self.properties.Get(SERVICE_IFACE, COLLECTIONS_PROPERTY_OLD,
                                reply_handler=propertyget_handler,
                                error_handler=d.errback)

        self.properties.Get(SERVICE_IFACE, COLLECTIONS_PROPERTY,
                            reply_handler=propertyget_handler,
                            error_handler=error_fallback)
        return d

    def get_default_collection(self):
        """The collection where default items should be created."""
        d = Deferred()

        def prompt_handle(unlocked):
            """Handle showing a prompt."""
            collection_path = unlocked[0]
            return Collection(self, collection_path)

        def unlock_handler(unlocked, prompt):
            """The objects were unlocked."""
            if prompt != "/":
                d2 = self.do_prompt(prompt)
                d2.addCallback(prompt_handle)
                d2.chainDeferred(d)
            else:
                d.callback(prompt_handle(unlocked))

        def set_default_alias(collection):
            """Set the newly created collection as the default one."""
            d4 = Deferred()
            alias_set = lambda: d4.callback(collection)
            object_path = dbus.ObjectPath(collection.object_path)
            self.service.SetAlias(DEFAULT_LABEL, object_path,
                                  reply_handler=alias_set,
                                  error_handler=d4.errback)
            return d4

        def readalias_handler(collection_path):
            """ReadAlias returned."""
            if collection_path != "/":
                # The collection was found, make sure it's unlocked
                objects = dbus.Array([collection_path], signature="o")
                self.service.Unlock(objects,
                                    reply_handler=unlock_handler,
                                    error_handler=d.errback)
            else:
                # The collection was not found, so create it
                d3 = self.create_collection(DEFAULT_LABEL)
                d3.addCallback(set_default_alias)
                d3.chainDeferred(d)

        def default_collection_not_found(e):
            """Try the default alias."""
            self.service.ReadAlias(DEFAULT_LABEL,
                                   reply_handler=readalias_handler,
                                   error_handler=d.errback)

        def found_default_collection(label):
            """Make sure the default collection is unlocked."""
            objects = dbus.Array([DEFAULT_COLLECTION], signature="o")
            self.service.Unlock(objects,
                                reply_handler=unlock_handler,
                                error_handler=d.errback)

        collection = Collection(self, DEFAULT_COLLECTION)
        d0 = collection.get_label()
        d0.addCallback(found_default_collection)
        d0.addErrback(default_collection_not_found)

        return d


class Collection(object):
    """A collection of items containing secrets."""

    def __init__(self, service, object_path):
        """Initialize a new collection."""
        self.service = service
        self.object_path = object_path
        collection_object = service.bus.get_object(BUS_NAME, object_path,
                                                   introspect=False)
        self.collection_iface = dbus.Interface(collection_object,
                                               dbus_interface=COLLECTION_IFACE)
        self.properties = dbus.Interface(collection_object,
                                         dbus_interface=PROPERTIES_IFACE)

    def get_label(self):
        """Return the label for this collection from the keyring."""
        d = Deferred()

        def error_fallback(error):
            """Fall back to the old property name."""
            self.properties.Get(COLLECTION_IFACE, CLXN_LABEL_PROPERTY_OLD,
                                reply_handler=d.callback,
                                error_handler=d.errback)

        self.properties.Get(COLLECTION_IFACE, CLXN_LABEL_PROPERTY,
                            reply_handler=d.callback,
                            error_handler=error_fallback)
        return d

    def create_item(self, label, attr, value, replace=True):
        """Create an item with the given attributes, secret and label.

        If replace is set, then it replaces an item already present with the
        same values for the attributes.
        """
        d = Deferred()

        def createitem_handler(item, prompt):
            """An item was created, or a prompt should be shown first."""
            if prompt != "/":
                self.service.do_prompt(prompt).chainDeferred(d)
            else:
                d.callback(item)

        properties = dbus.Dictionary(signature="sv")
        properties[ITEM_LABEL_PROPERTY] = label
        attributes = dbus.Dictionary(attr, signature="ss")
        properties[ITEM_ATTRIBUTES_PROPERTY] = attributes
        parameters = dbus.ByteArray(ALGORITHM_PARAMS)
        value_bytes = dbus.ByteArray(value)
        secret = (self.service.session, parameters, value_bytes,
                  SECRET_CONTENT_TYPE)

        def error_fallback(error):
            """A fallback for using old property names and signature."""
            oldprops = dbus.Dictionary(signature="sv")
            oldprops[ITEM_LABEL_PROPERTY_OLD] = label
            oldprops[ITEM_ATTRIBUTES_PROPERTY_OLD] = attributes
            secret = (self.service.session, parameters, value_bytes)
            self.collection_iface.CreateItem(oldprops, secret, replace,
                                             reply_handler=createitem_handler,
                                             error_handler=d.errback)

        self.collection_iface.CreateItem(properties, secret, replace,
                                         reply_handler=createitem_handler,
                                         error_handler=error_fallback)
        return d


class Item(object):
    """An item contains a secret, lookup attributes and has a label."""

    def __init__(self, service, object_path):
        """Initialize this new Item."""
        self.service = service
        self.object_path = object_path
        item_object = service.bus.get_object(BUS_NAME, object_path)
        self.item_iface = dbus.Interface(item_object,
                                         dbus_interface=ITEM_IFACE)

    def get_value(self):
        """Retrieve the secret for this item."""
        d = Deferred()

        def getsecret_handler(secret):
            """The secret for this item was found."""
            # pylint: disable=W0612
            value = secret[2]
            d.callback(value)

        self.item_iface.GetSecret(self.service.session, byte_arrays=True,
                                  reply_handler=getsecret_handler,
                                  error_handler=d.errback)
        return d

    def delete(self):
        """Delete this item."""
        d = Deferred()

        def delete_handler(prompt):
            """The item was deleted, or a prompt should be shown first."""
            if prompt != "/":
                self.service.do_prompt(prompt).chainDeferred(d)
            else:
                d.callback(True)

        self.item_iface.Delete(reply_handler=delete_handler,
                               error_handler=d.errback)
        return d



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the main module."""

from copy import copy
import os
import sys

from PyQt4 import QtCore
from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.qt import main
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso import tests


# pylint: disable=C0103
class FakeUi(object):

    """A fake UI."""

    def size(self):
        """Fake size."""
        return QtCore.QSize(100, 100)

    def setGeometry(self, *args):
        """Fake setGeometry."""

    show = setGeometry
    raise_ = lambda self: None


class FakeDesktop(object):

    """Fake Desktop Widget."""

    def availableGeometry(self):
        """Fake availableGeometry for desktop.-"""
        return QtCore.QRect(100, 100, 100, 100)


class FakeApplication(object):

    """Fake QApplication."""

    called = {}
    __instance = None

    def __init__(self, args):
        self.called['args'] = args
        FakeApplication.__instance = self

    def setStyleSheet(self, style):
        """Fake setStyleSheet."""
        self.called["setStyleSheet"] = style

    def styleSheet(self):
        """Fake get style sheet."""
        return self.called.get("setStyleSheet", '')

    def desktop(self):
        """Fake Desktop."""
        return FakeDesktop()

    def exec_(self):
        """Fake exec_."""

    def exit(self):
        """Fake exit."""

    def setWindowIcon(self, icon_path):
        """Fake setWindowIcon."""

    @classmethod
    def instance(cls):
        """Fake instance."""
        return FakeApplication.__instance
# pylint: enable=C0103


class BasicTestCase(TestCase):
    """Test case with a helper tracker."""

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BasicTestCase, self).setUp()
        self.called = []
        self._close_callback = None
        self._app = None

        def called_ui(method, *args, **kw):
            """record ui call."""
            if 'close_callback' in kw:
                self._close_callback = kw.pop('close_callback')
            self.called.append((method, args, kw))
            return FakeUi()

        self.patch(main, 'UbuntuSSOClientGUI',
            lambda *arg, **kw: called_ui('GUI', *arg, **kw))
        self.patch(main.source, 'main',
            lambda *arg, **kw: called_ui('main', *arg, **kw))
        self.patch(main.source, 'main_start',
            lambda *arg, **kw: called_ui('main_start', *arg, **kw))
        self.patch(main.QtGui, 'QApplication', FakeApplication)
        FakeApplication.called['args'] = []

    def test_main(self):
        """Calling main.main() a UI instance is created."""
        kwargs = dict(app_name='APP_NAME', foo='foo', bar='bar',
            baz='yadda', yadda=0)
        main.main(**kwargs)

        expected = [('main', (FakeApplication.instance(),), {}),
            ('GUI', (), kwargs),
            ('main_start', (FakeApplication.instance(),), {})]
        self.assertEqual(self.called, expected)

    def test_testability(self):
        """Calling main.main() with TESTABILITY=1, adds -testability."""
        environ = copy(os.environ)
        environ['TESTABILITY'] = '1'
        self.patch(os, 'environ', environ)
        self.patch(sys, 'argv', [])
        kwargs = dict(app_name='APP_NAME', foo='foo', bar='bar',
            baz='yadda', yadda=0)
        main.main(**kwargs)
        self.assertIn('-testability', FakeApplication.called['args'])

    def test_no_testability(self):
        """Calling main.main() with TESTABILITY unset."""
        environ = copy(os.environ)
        if 'TESTABILITY' in environ:
            environ['TESTABILITY'] = ''
        self.patch(os, 'environ', environ)
        self.patch(sys, 'argv', [])
        kwargs = dict(app_name='APP_NAME', foo='foo', bar='bar',
            baz='yadda', yadda=0)
        main.main(**kwargs)
        self.assertNotIn('-testability', FakeApplication.called['args'])

    def test_check_close_callback(self):
        """Check that the close callback is main_quit."""

        def called_quit(app):
            """Record the call to quit."""
            self._app = app

        self.patch(main.source, 'main_quit', called_quit)

        kwargs = dict(app_name='APP_NAME', foo='foo', bar='bar',
            baz='yadda', yadda=0)
        main.main(**kwargs)

        expected = [('main', (FakeApplication.instance(),), {}),
            ('GUI', (), kwargs),
            ('main_start', (FakeApplication.instance(),), {})]
        self._close_callback()
        self.assertEqual(self.called, expected)
        self.assertEqual(self._app, FakeApplication.instance())

    def test_main_args(self):
        """Calling main.main() a UI instance is created."""
        arg_list = (tests.APP_NAME, tests.NAME, tests.PASSWORD,
            tests.EMAIL_TOKEN)
        kwargs = dict(app_name=arg_list[0].encode('utf-8'),
            foo=arg_list[1].encode('utf-8'),
            bar=arg_list[2].encode('utf-8'),
            baz=arg_list[3].encode('utf-8'))
        main.main(**kwargs)

        args_unicode = dict(app_name=arg_list[0], foo=arg_list[1],
            bar=arg_list[2], baz=arg_list[3])
        expected = [('main', (FakeApplication.instance(),), {}),
            ('GUI', (), args_unicode),
            ('main_start', (FakeApplication.instance(),), {})]
        self.assertEqual(self.called, expected)

    def test_styles_load(self):
        """Test that all stylesheets load."""
        kwargs = dict(foo='foo', bar='bar', baz='yadda', yadda=0)
        main.main(**kwargs)
        data = []
        for qss_name in (main.PLATFORM_QSS, ":/stylesheet.qss"):
            qss = QtCore.QResource(qss_name)
            data.append(compat.text_type(qss.data()))
        self.assertEqual(
            compat.text_type(main.QtGui.QApplication.instance().styleSheet()),
            '\n'.join(data))



# -*- coding: utf-8 -*-

# Authors: Diego Sarmentero <diego.sarmentero@canonical.com>
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Customized Check Box to support links."""

from PyQt4 import QtGui, QtCore


class EnhancedCheckBox(QtGui.QCheckBox):
    """Enhanced QCheckBox to support links in the message displayed."""

    def __init__(self, text="", parent=None):
        QtGui.QCheckBox.__init__(self, parent)
        hbox = QtGui.QHBoxLayout()
        hbox.setAlignment(QtCore.Qt.AlignLeft)
        self.text_label = QtGui.QLabel(text)
        self.text_label.setWordWrap(True)
        self.text_label.setOpenExternalLinks(True)
        padding = self.iconSize().width()
        self.text_label.setStyleSheet("margin-top: -3px;"
                                      "padding-left: 2px;")
        hbox.setContentsMargins(padding, 0, 0, 0)
        hbox.addWidget(self.text_label)
        self.setLayout(hbox)

        if parent is not None:
            lines = self.text_label.width() / float(parent.width())
            self.text_label.setMinimumWidth(parent.width())
            self.setMinimumHeight(self.height() * lines)

        self.stateChanged.connect(self.text_label.setFocus)

    def text(self):
        """Return the text of this widget."""
        return self.text_label.text()

    # Invalid name "setText"
    # pylint: disable=C0103

    def setText(self, text):
        """Set a new text to this widget."""
        self.text_label.setText(text)

    # pylint: enable=C0103



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright (C) 2010-2012 Canonical
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Handle keys in the local kerying."""

try:
    # pylint: disable=E0611,F0401
    from urllib.parse import parse_qsl, urlencode
    # pylint: enable=E0611,F0401
except ImportError:
    from urllib import urlencode
    from urlparse import parse_qsl

from twisted.internet.defer import inlineCallbacks, returnValue

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils.txsecrets import SecretService
from ubuntu_kylin_sso.keyring import (
    get_token_name,
    get_old_token_name,
    U1_APP_NAME,
    try_old_credentials)


logger = setup_logging("ubuntu_kylin_sso.keyring.linux")


class Keyring(object):
    """A Keyring for a given application name."""

    def __init__(self):
        """Initialize this instance."""
        self.service = SecretService()

    @inlineCallbacks
    def _find_keyring_item(self, app_name, attr=None):
        """Return the keyring item or None if not found."""
        if attr is None:
            attr = self._get_keyring_attr(app_name)
        logger.debug("Finding all items for app_name %r.", app_name)
        items = yield self.service.search_items(attr)
        if len(items) == 0:
            # if no items found, return None
            logger.debug("No items found!")
            returnValue(None)

        logger.debug("Returning first item found.")
        returnValue(items[0])

    def _get_keyring_attr(self, app_name):
        """Build the keyring attributes for this credentials."""
        attr = {"key-type": "Ubuntu Kylin SSO credentials",
                "token-name": get_token_name(app_name)}
        return attr

    @inlineCallbacks
    def set_credentials(self, app_name, cred):
        """Set the credentials of the Ubuntu SSO item."""
        # Creates the secret from the credentials
        secret = urlencode(cred)

        attr = self._get_keyring_attr(app_name)
        # Add our SSO credentials to the keyring
        yield self.service.open_session()
        collection = yield self.service.get_default_collection()
        yield collection.create_item(app_name, attr, secret, True)

    @inlineCallbacks
    def _migrate_old_token_name(self, app_name):
        """Migrate credentials with old name, store them with new name."""
        logger.debug("Migrating old token name.")
        attr = self._get_keyring_attr(app_name)
        attr['token-name'] = get_old_token_name(app_name)
        item = yield self._find_keyring_item(app_name, attr=attr)
        if item is not None:
            yield self.set_credentials(app_name,
                                       dict(parse_qsl(item.secret)))
            yield item.delete()

        result = yield self._find_keyring_item(app_name)
        returnValue(result)

    @inlineCallbacks
    def get_credentials(self, app_name):
        """A deferred with the secret of the SSO item in a dictionary."""
        # If we have no attributes, return None
        logger.debug("Getting credentials for %r.", app_name)
        yield self.service.open_session()
        item = yield self._find_keyring_item(app_name)
        if item is None:
            item = yield self._migrate_old_token_name(app_name)

        if item is not None:
            logger.debug("Parsing secret.")
            secret = yield item.get_value()
            returnValue(dict(parse_qsl(secret)))
        # Robert 20140825: Special code for Ubuntu One old keyring. Do not use it in UKID.
        # else:
        #     # if no item found, try getting the old credentials
        #     if app_name == U1_APP_NAME:
        #         logger.debug("Trying old credentials for %r.", app_name)
        #         old_creds = yield try_old_credentials(app_name)
        #         returnValue(old_creds)
        # nothing was found
        returnValue(None)

    @inlineCallbacks
    def delete_credentials(self, app_name):
        """Delete a set of credentials from the keyring."""
        attr = self._get_keyring_attr(app_name)
        # Add our SSO credentials to the keyring
        yield self.service.open_session()
        collection = yield self.service.get_default_collection()
        yield collection.create_item(app_name, attr, "secret!", True)

        item = yield self._find_keyring_item(app_name)
        if item is not None:
            yield item.delete()



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Script that shows the qt gui."""

from __future__ import unicode_literals

import sys

# pylint: disable=E1101,F0401,E0611

if sys.platform != 'win32':
    from twisted.internet import gireactor
    gireactor.install()

    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)

from twisted.internet import defer, reactor

from ubuntu_kylin_sso.main import get_sso_client
from ubuntu_kylin_sso.credentials import (
    HELP_TEXT_KEY,
    PING_URL_KEY,
    POLICY_URL_KEY,
    TC_URL_KEY,
)


@defer.inlineCallbacks
def main():
    """Perform a client request to be logged in."""
    d = defer.Deferred()
    client = yield get_sso_client()

    def found(*args):
        """The result was received."""
        d.callback(args)

    client.cred_manager.connect_to_signal('CredentialsFound', found)
    client.cred_manager.connect_to_signal('CredentialsNotFound', found)
    client.cred_manager.connect_to_signal('AuthorizationDenied', found)
    client.cred_manager.connect_to_signal('CredentialsError', found)

    yield client.cred_manager.login('Ubuntu Two',
        {
            HELP_TEXT_KEY: 'This is a test help text.',
            PING_URL_KEY: 'http://www.wikipedia.com/',
            POLICY_URL_KEY: 'http://one.ubuntu.com/',
            TC_URL_KEY: 'http://www.google.com/',
        })
    yield d

    yield client.disconnect()
    reactor.stop()


if __name__ == '__main__':
    main()
    reactor.run()



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Single Sign On client main module.

Provides a utility which accepts requests to the Ubuntu Single Sign On
service. The OAuth process is handled, including adding the OAuth access token
to the local keyring.

"""

import sys

from twisted.internet import defer

from ubuntu_kylin_sso import utils
from ubuntu_kylin_sso.account import Account
from ubuntu_kylin_sso.credentials import (
    Credentials,
    HELP_TEXT_KEY,
    PING_URL_KEY,
    POLICY_URL_KEY,
    TC_URL_KEY,
    UI_EXECUTABLE_KEY,
    UserCancellationError,
    WINDOW_ID_KEY,
)
from ubuntu_kylin_sso.keyring import get_token_name, Keyring
from ubuntu_kylin_sso.logger import setup_logging, log_call
from ubuntu_kylin_sso.utils import compat


logger = setup_logging("ubuntu_kylin_sso.main")
TIMEOUT_INTERVAL = 10000  # 10 seconds

# pylint: disable=C0103, W0703, W0621

if sys.platform == 'win32':
    from ubuntu_kylin_sso.main import windows as source
    from ubuntu_kylin_sso.main.perspective_broker import (
        finish_setup,
        main as main_func,
        timeout_func,
        shutdown_func,
        start_setup,
    )

    source.finish_setup = finish_setup
    source.timeout_func = timeout_func
    source.shutdown_func = shutdown_func
    source.start_setup = start_setup
    source.main = main_func

    TIMEOUT_INTERVAL = 10000000000  # forever (hack)
elif sys.platform == 'darwin':
    from ubuntu_kylin_sso.main import darwin as source
    from ubuntu_kylin_sso.main.perspective_broker import (
        finish_setup,
        main as main_func,
        timeout_func,
        shutdown_func,
        start_setup,
    )

    source.finish_setup = finish_setup
    source.timeout_func = timeout_func
    source.shutdown_func = shutdown_func
    source.start_setup = start_setup
    source.main = main_func
else:
    from ubuntu_kylin_sso.main import linux as source


UbuntuSSOClient = source.UbuntuSSOClient
UbuntuSSOProxy = source.UbuntuSSOProxy
get_sso_client = source.get_sso_client


def except_to_errdict(e):
    """Turn an exception into a dictionary to return thru IPC."""
    result = {
        "errtype": e.__class__.__name__,
    }
    if len(e.args) == 0:
        result["message"] = e.__class__.__doc__
    elif isinstance(e.args[0], dict):
        result.update(e.args[0])
    elif isinstance(e.args[0], compat.basestring):
        result["message"] = e.args[0]

    return result


class SSOLogin(object):
    """Login thru the Single Sign On service."""

    def __init__(self, proxy):
        """Initiate the Login object."""
        self.processor = Account()
        self.proxy = proxy

    @defer.inlineCallbacks
    def _process_new_token(self, app_name, email, credentials, ping_url):
        """Process a new set of credentials for 'email'."""
        if ping_url:
            yield utils.ping_url(ping_url, email, credentials)
            logger.info('Url %r successfully opened!', ping_url)

        yield Keyring().set_credentials(app_name, credentials)

    @log_call(logger.debug)
    def CaptchaGenerated(self, app_name, result):
        """Signal thrown after the captcha is generated."""
        self.proxy.CaptchaGenerated(app_name, result)

    @log_call(logger.debug)
    def CaptchaGenerationError(self, app_name, error):
        """Signal thrown when there's a problem generating the captcha."""
        error_dict = except_to_errdict(error)
        self.proxy.CaptchaGenerationError(app_name, error_dict)

    @defer.inlineCallbacks
    def generate_captcha(self, app_name, filename):
        """Call the matching method in the processor."""
        try:
            result = yield self.processor.generate_captcha(filename)
            self.CaptchaGenerated(app_name, result)
        except Exception as e:
            self.CaptchaGenerationError(app_name, e)

    @log_call(logger.debug)
    def UserRegistered(self, app_name, result):
        """Signal thrown when the user is registered."""
        self.proxy.UserRegistered(app_name, result)

    @log_call(logger.debug)
    def UserRegistrationError(self, app_name, error):
        """Signal thrown when there's a problem registering the user."""
        error_dict = except_to_errdict(error)
        self.proxy.UserRegistrationError(app_name, error_dict)

    @defer.inlineCallbacks
    def register_user(self, app_name, email, password, name, captcha_id,
                      captcha_solution):
        """Call the matching method in the processor."""
        try:
            result = yield self.processor.register_user(email, password, name,
                                                captcha_id, captcha_solution)
            self.UserRegistered(app_name, result)
        except Exception as e:
            self.UserRegistrationError(app_name, e)

    @log_call(logger.debug)
    def LoggedIn(self, app_name, result):
        """Signal thrown when the user is logged in."""
        self.proxy.LoggedIn(app_name, result)

    @log_call(logger.debug)
    def LoginError(self, app_name, error):
        """Signal thrown when there is a problem in the login."""
        error_dict = except_to_errdict(error)
        self.proxy.LoginError(app_name, error_dict)

    @log_call(logger.debug)
    def UserNotValidated(self, app_name, email):
        """Signal thrown when the user is not validated."""
        self.proxy.UserNotValidated(app_name, email)

    @defer.inlineCallbacks
    def login(self, app_name, email, password, ping_url=None):
        """Call the matching method in the processor."""
        try:
            token_name = get_token_name(app_name)
            logger.debug('login: token_name %r, email %r, password <hidden>.',
                         token_name, email)
            credentials = yield self.processor.login(email, password,
                                                     token_name)
            logger.debug('login returned not None credentials? %r.',
                         credentials is not None)
            is_validated = yield self.processor.is_validated(credentials)
            logger.debug('user is validated? %r.', is_validated)
            if is_validated:
                yield self._process_new_token(app_name, email,
                                              credentials, ping_url)
                self.LoggedIn(app_name, email)
            else:
                self.UserNotValidated(app_name, email)
        except Exception as e:
            self.LoginError(app_name, e)

    @log_call(logger.debug)
    def EmailValidated(self, app_name, result):
        """Signal thrown after the email is validated."""
        self.proxy.EmailValidated(app_name, result)

    @log_call(logger.debug)
    def EmailValidationError(self, app_name, error):
        """Signal thrown when there's a problem validating the email."""
        error_dict = except_to_errdict(error)
        self.proxy.EmailValidationError(app_name, error_dict)

    @defer.inlineCallbacks
    def validate_email(self, app_name, email, password, email_token,
                       ping_url=None):
        """Call the matching method in the processor."""
        try:
            token_name = get_token_name(app_name)
            credentials = yield self.processor.validate_email(email, password,
                                                      email_token, token_name)
            yield self._process_new_token(app_name, email,
                                          credentials, ping_url)
            self.EmailValidated(app_name, email)
        except Exception as e:
            self.EmailValidationError(app_name, e)

    @log_call(logger.debug)
    def PasswordResetTokenSent(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""
        self.proxy.PasswordResetTokenSent(app_name, result)

    @log_call(logger.debug)
    def PasswordResetError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""
        error_dict = except_to_errdict(error)
        self.proxy.PasswordResetError(app_name, error_dict)

    @defer.inlineCallbacks
    def request_password_reset_token(self, app_name, email):
        """Call the matching method in the processor."""
        try:
            result = yield self.processor.request_password_reset_token(email)
            self.PasswordResetTokenSent(app_name, result)
        except Exception as e:
            self.PasswordResetError(app_name, e)

    @log_call(logger.debug)
    def PasswordChanged(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""
        self.proxy.PasswordChanged(app_name, result)

    @log_call(logger.debug)
    def PasswordChangeError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""
        error_dict = except_to_errdict(error)
        self.proxy.PasswordChangeError(app_name, error_dict)

    @defer.inlineCallbacks
    def set_new_password(self, app_name, email, token, new_password):
        """Call the matching method in the processor."""
        try:
            result = yield self.processor.set_new_password(email, token,
                                                           new_password)
            self.PasswordChanged(app_name, result)
        except Exception as e:
            self.PasswordChangeError(app_name, e)


class CredentialsManagement(object):
    """Object that manages credentials.

    Every exposed method in this class requires one mandatory argument:

        - 'app_name': the name of the application. Will be displayed in the
        GUI header, plus it will be used to find/build/clear tokens.

    And accepts another parameter named 'args', which is a dictionary that
    can contain the following:

        - 'help_text': an explanatory text for the end-users, will be
        shown below the header. This is an optional free text field.

        - 'ping_url': the url to open after successful token retrieval. If
        defined, the email will be attached to the url and will be pinged
        with a OAuth-signed request.

        - 'tc_url': the link to the Terms and Conditions page. If defined,
        the checkbox to agree to the terms will link to it.

        - 'window_id': the id of the window which will be set as a parent
        of the GUI. If not defined, no parent will be set.

    """

    def __init__(self, timeout_func, shutdown_func, proxy):
        super(CredentialsManagement, self).__init__()
        self._ref_count = 0
        self.timeout_func = timeout_func
        self.shutdown_func = shutdown_func
        self.proxy = proxy

    def _get_ref_count(self):
        """Get value of ref_count."""
        return self._ref_count

    def _set_ref_count(self, new_value):
        """Set a new value to ref_count."""
        logger.debug('ref_count is %r, changing value to %r.',
                     self._ref_count, new_value)
        if new_value < 0:
            self._ref_count = 0
            msg = 'Attempting to decrease ref_count to a negative value (%r).'
            logger.warning(msg, new_value)
        else:
            self._ref_count = new_value

        if self._ref_count == 0:
            logger.debug('Setting up timer with %r (%r, %r).',
                         self.timeout_func, TIMEOUT_INTERVAL, self.shutdown)
            self.timeout_func(TIMEOUT_INTERVAL, self.shutdown)

    ref_count = property(fget=_get_ref_count, fset=_set_ref_count)

    def shutdown(self):
        """If no ongoing requests, call self.shutdown_func."""
        logger.debug('shutdown!, ref_count is %r.', self._ref_count)
        if self._ref_count == 0:
            logger.info('Shutting down, calling %r.', self.shutdown_func)
            self.shutdown_func()

    valid_keys = (HELP_TEXT_KEY, PING_URL_KEY, POLICY_URL_KEY, TC_URL_KEY,
                  UI_EXECUTABLE_KEY, WINDOW_ID_KEY)

    def _parse_args(self, args):
        """Retrieve values from the generic param 'args'."""
        result = dict(i for i in args.items() if i[0] in self.valid_keys)
        result[WINDOW_ID_KEY] = int(args.get(WINDOW_ID_KEY, 0))
        return result

    @log_call(logger.info)
    def AuthorizationDenied(self, app_name):
        """Signal thrown when the user denies the authorization."""
        self.ref_count -= 1
        self.proxy.AuthorizationDenied(app_name)

    # do not use log_call decorator since we should not log credentials
    def CredentialsFound(self, app_name, credentials):
        """Signal thrown when the credentials are found."""
        self.ref_count -= 1
        logger.info('%s: emitting CredentialsFound with app_name %r.',
                    self.__class__.__name__, app_name)
        self.proxy.CredentialsFound(app_name, credentials)

    @log_call(logger.info)
    def CredentialsNotFound(self, app_name):
        """Signal thrown when the credentials are not found."""
        self.ref_count -= 1
        self.proxy.CredentialsNotFound(app_name)

    @log_call(logger.info)
    def CredentialsCleared(self, app_name):
        """Signal thrown when the credentials were cleared."""
        self.ref_count -= 1
        self.proxy.CredentialsCleared(app_name)

    @log_call(logger.info)
    def CredentialsStored(self, app_name):
        """Signal thrown when the credentials were cleared."""
        self.ref_count -= 1
        self.proxy.CredentialsStored(app_name)

    @log_call(logger.error)
    def CredentialsError(self, app_name, error):
        """Signal thrown when there is a problem getting the credentials."""
        self.ref_count -= 1
        if isinstance(error, dict):
            error_dict = error
        else:
            error_dict = except_to_errdict(error)
        self.proxy.CredentialsError(app_name, error_dict)

    def find_credentials(self, app_name, args, success_cb=None, error_cb=None):
        """Look for the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        - 'success_cb', if not None, will be executed if the operation was
        a success.

        - 'error_cb', if not None, will be executed if the operation had
        an error.

        """
        def _analize_creds(credentials):
            """Find credentials and notify using signals."""
            if credentials is not None and len(credentials) > 0:
                self.CredentialsFound(app_name, credentials)
            else:
                self.CredentialsNotFound(app_name)

        def _tweaked_success_cb(creds):
            """Decrease ref counter and call 'success_cb'."""
            self.ref_count -= 1
            success_cb(creds)

        if success_cb is None:
            _success_cb = _analize_creds
        else:
            _success_cb = _tweaked_success_cb

        def _tweaked_error_cb(error, app):
            """Decrease ref counter and call 'error_cb', modifying the dict."""
            self.ref_count -= 1
            error_cb(except_to_errdict(error.value))

        if error_cb is None:
            _error_cb = lambda f, _: self.CredentialsError(app_name, f.value)
        else:
            _error_cb = _tweaked_error_cb

        self.ref_count += 1
        obj = Credentials(app_name)
        d = obj.find_credentials()
        d.addCallback(_success_cb)
        d.addErrback(_error_cb, app_name)

    def _process_failures(self, failure, app_name):
        """Process failure returned by the Credentials module."""
        if failure.check(UserCancellationError):
            self.AuthorizationDenied(app_name)
        else:
            self.CredentialsError(app_name, failure.value)

    def clear_credentials(self, app_name, args):
        """Clear the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        """
        self.ref_count += 1
        obj = Credentials(app_name)
        d = obj.clear_credentials()
        d.addCallback(lambda _: self.CredentialsCleared(app_name))
        d.addErrback(lambda f: self.CredentialsError(app_name, f.value))

    def store_credentials(self, app_name, args):
        """Store the token for an application.

        - 'app_name': the name of the application which credentials are
        going to be stored.

        - 'args' is the dictionary holding the credentials. Needs to provide
        the following mandatory keys: 'token', 'token_key', 'consumer_key',
        'consumer_secret'.

        """
        self.ref_count += 1
        obj = Credentials(app_name)
        d = obj.store_credentials(args)
        d.addCallback(lambda _: self.CredentialsStored(app_name))
        d.addErrback(lambda f: self.CredentialsError(app_name, f.value))

    def register(self, app_name, args):
        """Get credentials if found else prompt GUI to register."""
        self.ref_count += 1
        obj = Credentials(app_name, **self._parse_args(args))
        d = obj.register()
        d.addCallback(lambda creds: self.CredentialsFound(app_name, creds))
        d.addErrback(self._process_failures, app_name)

    def login(self, app_name, args):
        """Get credentials if found else prompt GUI to login."""
        self.ref_count += 1
        obj = Credentials(app_name, **self._parse_args(args))
        d = obj.login()
        d.addCallback(lambda creds: self.CredentialsFound(app_name, creds))
        d.addErrback(self._process_failures, app_name)

    def login_email_password(self, app_name, args):
        """Get credentials if found else try to login.

        Login will be done by inspecting 'args' and expecting to find two keys:
        'email' and 'password'.

        """
        self.ref_count += 1
        email = args.pop('email')
        password = args.pop('password')
        obj = Credentials(app_name, **self._parse_args(args))
        d = obj.login(email=email, password=password)
        d.addCallback(lambda creds: self.CredentialsFound(app_name, creds))
        d.addErrback(self._process_failures, app_name)


# pylint: enable=C0103

class UbuntuSSOService(object):
    """Manager that exposes the diff referenceable objects."""

    def __init__(self):
        self.proxy = UbuntuSSOProxy(self)
        self.sso_login = None
        self.cred_manager = None

    @defer.inlineCallbacks
    def start(self):
        """Start the service."""
        logger.debug('Starting up Ubuntu SSO service...')
        try:
            yield self.proxy.start()
        except:
            logger.exception('Can not start Ubuntu SSO service:')
            raise
        else:
            logger.info('Ubuntu SSO service started.')

        self.sso_login = SSOLogin(proxy=self.proxy.sso_login)
        self.cred_manager = CredentialsManagement(
                                timeout_func=source.timeout_func,
                                shutdown_func=source.shutdown_func,
                                proxy=self.proxy.cred_manager)

    def shutdown(self):
        """Shutdown the service."""
        return self.proxy.shutdown()


def main():
    """Run the backend service."""
    logger.info('Setting up Ubuntu SSO service.')
    loop = source.start_setup()
    service = UbuntuSSOService()
    d = service.start()
    d.addBoth(source.finish_setup, loop)
    source.main()



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the main SSO Linux client code."""

# pylint: disable=C0103
do_tests = False
try:
    import dbus
    import dbus.service
    from ubuntu_kylin_sso.main import linux as main
    do_tests = True
except ImportError:
    do_tests = False
# pylint enable=C0103

from ubuntuone.devtools.testcases import BaseTestCase, skipIf


@skipIf(not do_tests, 'Tests only work with DBus and linux imports.')
class LinuxProxyTestCase(BaseTestCase):
    """Test some Linux specific cases."""

    def test_dbus_missing_exists_clean(self):
        """Test that dbus-daemon not being available gives us a clean exit."""
        def patched(*args, **kwargs):
            """Method to patch in for testing."""
            raise dbus.service.DBusException(
                'org.freedesktop.DBus.Error.NoServer')

        self.patch(dbus, "SessionBus", patched)
        self.patch(main, "shutdown_func", patched)
        self.assertRaises(dbus.service.DBusException,
                          main.UbuntuSSOProxy, None)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the glib runner helper module."""

import subprocess

# pylint: disable=E0611
from gi.repository import GLib
# pylint: enable=E0611
from twisted.internet import defer

from ubuntu_kylin_sso.utils import runner
from ubuntu_kylin_sso.utils.runner import glib
from ubuntu_kylin_sso.utils.runner.tests.test_runner import SpawnProgramTestCase


class FakedSignal(object):
    """Fake a glib signal."""

    def __init__(self, name):
        self.name = name
        self._handlers = []

    def connect(self, handler):
        """Connect 'handler' with this signal."""
        self._handlers.append(handler)

    def emit(self, *args, **kwargs):
        """Emit this signal."""
        for handler in self._handlers:
            handler(*args, **kwargs)


class FakedProcess(object):
    """Fake a glib Process."""

    _argv = _flags = None
    _pid = 123456
    _pids = []
    _status_code = 0

    SpawnFlags = GLib.SpawnFlags
    GError = GLib.GError

    def __init__(self):
        self.pid = lambda: self._pid
        self.started = FakedSignal('started')
        self.finished = FakedSignal('finished')
        self.error = FakedSignal('error')

    def spawn_async(self, argv, flags):
        """Spwan a process."""
        if 'None' in argv:
            exc = GLib.GError()
            exc.message = str('Can not handle None')
            exc.code = 24
            raise exc

        self._argv = argv
        self._flags = flags

        try:
            subprocess.call(argv)
        except Exception as e:
            exc = GLib.GError()
            exc.message = str(e)
            exc.code = 42
            raise exc

        self._pids.append(self._pid)
        return (self._pid, None, None, None)

    def child_watch_add(self, pid, child_watch):
        """Addd a child watch."""
        if pid in self._pids:
            child_watch(pid, self._status_code)

    def spawn_close_pid(self, pid):
        """Close the 'pid'."""
        self._pids.remove(pid)


class GLibSpawnProgramTestCase(SpawnProgramTestCase):
    """The test suite for the spawn_program method (using GLib)."""

    use_reactor = False

    @defer.inlineCallbacks
    def setUp(self):
        yield super(GLibSpawnProgramTestCase, self).setUp()
        # Since we can't mix plan glib runner and the gireactor, we patch
        # GLib.spawn_async and fake the conditions so the glib runner is chosen
        self.process = FakedProcess()
        self.patch(glib, 'GLib', self.process)
        self.patch(runner, 'is_qt4_main_loop_installed', lambda: False)

    # Access to a protected member _flags, _argv of a client class
    # pylint: disable=W0212

    @defer.inlineCallbacks
    def test_flags_are_correct(self):
        """The flags are the correct ones."""
        yield self.spawn_fn(self.args)

        flags = GLib.SpawnFlags.DO_NOT_REAP_CHILD | \
                GLib.SpawnFlags.SEARCH_PATH | \
                GLib.SpawnFlags.STDOUT_TO_DEV_NULL | \
                GLib.SpawnFlags.STDERR_TO_DEV_NULL
        self.assertEqual(self.process._flags, flags)

    @defer.inlineCallbacks
    def test_argv_is_bytes(self):
        """The argv parameter is converted to bytes."""
        yield self.spawn_fn(self.args)

        bytes_args = [a.encode('utf-8') for a in self.args]
        self.assertEqual(self.process._argv, bytes_args)



# -*- coding: utf-8 -*-

# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for txkeyring."""

import logging
import uuid

import dbus.service

from twisted.internet.defer import inlineCallbacks, returnValue
from ubuntuone.devtools.testcases.dbus import DBusTestCase

from ubuntu_kylin_sso.utils import txsecrets

# DBus exported objects have different naming conventions than vanilla python
# pylint: disable=C0103

# Dear pylint, my clock says... 2010. Do you need a tutorial on decorators?
# pylint: disable=C0322

# pylint complains when things are a little too dynamic
# pylint: disable=E1101

# Use of super on an old style class
# pylint: disable=E1002


ERROR_CREATE_BUT_LOCKED = "Cannot create an item in a locked collection"
PROMPT_BASE_PATH = "/org/freedesktop/secrets/prompt"
SESSION_BASE_PATH = "/org/freedesktop/secrets/session"
COLLECTION_BASE_PATH = "/org/freedesktop/secrets/collection/"
SAMPLE_CONTENT_TYPE = "text/plain; charset=utf8"


class InvalidProperty(Exception):
    """An exception for when invalid properties are passed in."""


class SampleMiscException(Exception):
    """An exception that will be turned into a DBus Exception."""


class ItemMock(dbus.service.Object):
    """An item contains a secret, lookup attributes and has a label."""
    get_secret_fail = False
    delete_fail = False
    delete_prompt = False
    dismissed = False

    def __init__(self, collection, label, attributes, value, *args, **kwargs):
        """Initialize this instance."""
        # pylint: disable=E1002
        super(ItemMock, self).__init__(*args, **kwargs)
        self.collection = collection
        self.label = label
        self.attributes = attributes
        self.value = value

    @dbus.service.method(dbus_interface=txsecrets.ITEM_IFACE,
                         out_signature="o")
    def Delete(self):
        """Delete this item."""
        if self.delete_fail:
            raise SampleMiscException()
        self.collection.items.remove(self)
        if self.delete_prompt:
            prompt_path = create_object_path(PROMPT_BASE_PATH)
            prompt = self.dbus_publish(prompt_path, PromptMock,
                                       result="",
                                       dismissed=self.dismissed)
            return prompt
        else:
            return "/"

    @dbus.service.method(dbus_interface=txsecrets.ITEM_IFACE,
                         in_signature="o", out_signature="(oayay)")
    def GetSecret(self, session):
        """Retrieve the secret for this item."""
        if self.get_secret_fail:
            raise SampleMiscException()
        return (session, "", self.value)

    def matches(self, search_attr):
        """See if this item matches a given search."""
        for k, val in search_attr.items():
            if k not in self.attributes:
                return False
            if self.attributes[k] != val:
                return False
        return True


class PromptMock(dbus.service.Object):
    """A prompt necessary to complete an operation."""

    def __init__(self, dismissed=True,
                 result=dbus.String("", variant_level=1), *args, **kwargs):
        """Initialize this instance."""
        # pylint: disable=E1002
        super(PromptMock, self).__init__(*args, **kwargs)
        self.dismissed = dismissed
        self.result = result

    @dbus.service.method(dbus_interface=txsecrets.PROMPT_IFACE,
                         in_signature="s")
    def Prompt(self, window_id):
        """Perform the prompt."""
        self.Completed(self.dismissed, self.result)

    @dbus.service.signal(dbus_interface=txsecrets.PROMPT_IFACE,
                         signature="bv")
    def Completed(self, dismissed, result):
        """The prompt and operation completed."""


class BaseCollectionMock(dbus.service.Object):
    """Base collection of items containing secrets."""
    SUPPORTS_MULTIPLE_OBJECT_PATHS = True
    SUPPORTS_MULTIPLE_CONNECTIONS = True
    create_item_prompt = False
    dismissed = False
    create_item_fail = False
    locked = False
    unlock_prompts = False
    item_mock_class = ItemMock

    item_attrs_property = txsecrets.ITEM_ATTRIBUTES_PROPERTY_OLD
    item_label_property = txsecrets.ITEM_LABEL_PROPERTY_OLD
    clxn_label_property = txsecrets.CLXN_LABEL_PROPERTY_OLD

    def __init__(self, label, *args, **kwargs):
        """Initialize this instance."""
        # pylint: disable=E1002
        super(BaseCollectionMock, self).__init__(*args, **kwargs)
        self.items = []
        self.label = label

    def _create_item(self, properties, secret, replace):
        """Create an item with the given attributes, secret and label.

        If replace is set, then it replaces an item already present with the
        same values for the attributes.
        """
        if self.create_item_fail:
            raise SampleMiscException()
        if self.locked:
            raise SampleMiscException(ERROR_CREATE_BUT_LOCKED)
        attributes = properties[self.item_attrs_property]
        item_label = properties[self.item_label_property]
        value = secret[2]
        item_path = create_object_path(make_coll_path(self.label))
        item = self.dbus_publish(item_path, self.item_mock_class, self,
                                 item_label, attributes, value)
        self.items.append(item)
        if self.create_item_prompt:
            prompt_path = create_object_path(PROMPT_BASE_PATH)
            prompt = self.dbus_publish(prompt_path, PromptMock,
                                       result=item,
                                       dismissed=self.dismissed)
            return "/", prompt
        else:
            return item, "/"

    @dbus.service.method(dbus_interface=txsecrets.PROPERTIES_IFACE,
                         in_signature="ss", out_signature="v")
    def Get(self, interface, propname):
        """The only property implemented is Label."""
        if interface == txsecrets.COLLECTION_IFACE and \
                propname == self.clxn_label_property:
            return dbus.String(self.label)
        raise InvalidProperty("Invalid property: {}".format(propname))


class CollectionMock(BaseCollectionMock):
    """Collection of items containing secrets."""

    @dbus.service.method(dbus_interface=txsecrets.COLLECTION_IFACE,
                         in_signature="a{sv}(oayay)b", out_signature="oo",
                         byte_arrays=True)
    def CreateItem(self, properties, secret, replace):
        """Expose the _create_item method on DBus."""
        assert len(secret) == 3
        return self._create_item(properties, secret, replace)


class SessionMock(dbus.service.Object):
    """A session tracks state between the service and a client application."""

    @dbus.service.method(dbus_interface=txsecrets.SESSION_IFACE)
    def Close(self):
        """Close this session."""


def make_coll_path(label):
    """Make the path to a collection with its label"""
    return COLLECTION_BASE_PATH + label


class SecretServiceMock(dbus.service.Object):
    """The Secret Service manages all the sessions and collections."""
    create_collection_prompt = False
    create_collection_fail = False
    open_session_fail = False
    dismissed = False
    collection_mock_class = CollectionMock

    clxn_label_property = txsecrets.CLXN_LABEL_PROPERTY_OLD
    collections_property = txsecrets.COLLECTIONS_PROPERTY_OLD

    def __init__(self, *args, **kwargs):
        """Initialize this instance."""
        # pylint: disable=E1002
        super(SecretServiceMock, self).__init__(*args, **kwargs)
        self.sessions = {}
        self.collections = {}
        self.aliases = {}

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="sv", out_signature="vo")
    def OpenSession(self, algorithm, algorithm_parameters):
        """Open a unique session for the caller application."""
        if self.open_session_fail:
            raise SampleMiscException()
        session_path = create_object_path(SESSION_BASE_PATH)
        session = self.dbus_publish(session_path, SessionMock)
        self.sessions[session_path] = session
        return True, session

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="a{sv}", out_signature="oo")
    def CreateCollection(self, properties):
        """Create a new collection with the specified properties."""
        if self.create_collection_fail:
            raise SampleMiscException()
        label = str(properties[self.clxn_label_property])
        coll_path = make_coll_path(label)
        collection = self.dbus_publish(coll_path, self.collection_mock_class,
                                       label)
        self.collections[label] = collection

        if self.create_collection_prompt:
            prompt_path = create_object_path(PROMPT_BASE_PATH)
            prompt = self.dbus_publish(prompt_path, PromptMock,
                                       result=collection,
                                       dismissed=self.dismissed)
            return "/", prompt
        else:
            return collection, "/"

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="a{ss}", out_signature="aoao")
    def SearchItems(self, attributes):
        """Find items in any collection."""
        unlocked_items = []
        locked_items = []
        for c in self.collections.values():
            if c.locked:
                append_item = locked_items.append
            else:
                append_item = unlocked_items.append
            for i in c.items:
                if i.matches(attributes):
                    append_item(i)

        return unlocked_items, locked_items

    def unlock_objects(self, objects):
        """Unlock the objects or its containers."""
        for c in self.collections.values():
            for l in c.locations:
                path = l[1]
                if path in objects:
                    c.locked = False
            for i in c.items:
                for l in i.locations:
                    path = l[1]
                    if path in objects:
                        c.locked = False

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="ao", out_signature="aoo")
    def Unlock(self, objects):
        """Unlock the specified objects."""
        locked = []
        unlocked = []
        for c in self.collections.values():
            for i in c.items:
                path = i.__dbus_object_path__
                if path in objects:
                    if c.unlock_prompts:
                        locked.append(path)
                    else:
                        unlocked.append(path)
        if locked:
            prompt_path = create_object_path(PROMPT_BASE_PATH)
            self.unlock_objects(objects)
            prompt = self.dbus_publish(prompt_path, PromptMock,
                                       result=locked,
                                       dismissed=self.dismissed)
            return unlocked, prompt
        else:
            self.unlock_objects(objects)
            return objects, "/"

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="s", out_signature="o")
    def ReadAlias(self, name):
        """Get the collection with the given alias."""
        result = self.aliases.get(name, "/")
        return result

    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="so")
    def SetAlias(self, name, collection_path):
        """Setup a collection alias."""
        self.aliases[name] = collection_path

    @dbus.service.method(dbus_interface=txsecrets.PROPERTIES_IFACE,
                         in_signature="ss", out_signature="v")
    def Get(self, interface, propname):
        """The only property implemented is Collections."""
        if interface == txsecrets.SERVICE_IFACE and \
                propname == self.collections_property:
            coll_paths = [make_coll_path(l) for l in self.collections]
            return dbus.Array(coll_paths, signature="o", variant_level=1)
        raise InvalidProperty("Invalid property: {}".format(propname))


class AltItemMock(ItemMock):
    """The secret in this item has a content_type."""

    @dbus.service.method(dbus_interface=txsecrets.ITEM_IFACE,
                         in_signature="o", out_signature="(oayays)")
    def GetSecret2(self, session):
        """Retrieve the secret for this item."""
        if self.get_secret_fail:
            raise SampleMiscException()
        return (session, "", self.value, SAMPLE_CONTENT_TYPE)


class AltCollectionMock(BaseCollectionMock):
    """The secrets in this collection have a content_type field."""

    item_mock_class = AltItemMock

    item_attrs_property = txsecrets.ITEM_ATTRIBUTES_PROPERTY
    item_label_property = txsecrets.ITEM_LABEL_PROPERTY
    clxn_label_property = txsecrets.CLXN_LABEL_PROPERTY

    @dbus.service.method(dbus_interface=txsecrets.COLLECTION_IFACE,
                         in_signature="a{sv}(oayays)b", out_signature="oo",
                         byte_arrays=True)
    def CreateItem(self, properties, secret, replace):
        """Expose the _create_item method on DBus."""
        assert len(secret) == 4
        return self._create_item(properties, secret, replace)


class AltSecretServiceMock(SecretServiceMock):
    """The secrets in this service have a content_type field."""

    collection_mock_class = AltCollectionMock

    clxn_label_property = txsecrets.CLXN_LABEL_PROPERTY
    collections_property = txsecrets.COLLECTIONS_PROPERTY

    # pylint: disable=W0221
    @dbus.service.method(dbus_interface=txsecrets.SERVICE_IFACE,
                         in_signature="a{sv}s", out_signature="oo")
    def CreateCollection(self, properties, alias):
        """Create a new collection with the specified properties."""
        # pylint: disable=E1002
        collection, prompt = super(AltSecretServiceMock,
                                   self).CreateCollection(properties)
        self.SetAlias(alias, collection)
        return collection, prompt
    # pylint: enable=W0221


def create_object_path(base):
    """Create a random object path given a base path."""
    random = uuid.uuid4().hex
    return base + "/" + random


class TextFilter(object):
    """Prevents the logging of messages containing a given text."""

    def __init__(self, *args):
        """Initialize this filter."""
        super(TextFilter, self).__init__()
        self.filtered_strings = args

    def filter(self, record):
        """See if we need to filter a given log record."""
        return not any(s in record.msg for s in self.filtered_strings)


class BaseTestCase(DBusTestCase):
    """Base class for DBus tests."""
    timeout = 10
    secret_service_class = SecretServiceMock

    @inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        self.session_bus = dbus.SessionBus()
        self.mock_service = self.dbus_publish(txsecrets.SECRETS_SERVICE,
                                              self.secret_service_class)
        self.secretservice = txsecrets.SecretService()
        self.silence_dbus_logging()

    def silence_dbus_logging(self):
        """Silence the warnings printed by dbus that pollute test results."""
        logger = logging.getLogger('dbus.connection')
        logfilter = TextFilter("Unable to set arguments")
        logger.addFilter(logfilter)
        self.addCleanup(logger.removeFilter, logfilter)

    def dbus_publish(self, object_path, object_class, *args, **kwargs):
        """Create an object and publish it on the bus."""
        name = self.session_bus.request_name(txsecrets.BUS_NAME,
                                             dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
        self.assertNotEqual(name, dbus.bus.REQUEST_NAME_REPLY_EXISTS)
        mock_object = object_class(*args, object_path=object_path,
                                   conn=self.session_bus, **kwargs)
        self.addCleanup(mock_object.remove_from_connection,
                        connection=self.session_bus,
                        path=object_path)
        mock_object.dbus_publish = self.dbus_publish
        return mock_object

    @inlineCallbacks
    def create_sample_collection(self, label, make_alias=True,
                                 publish_default_path=False):
        """Create a collection with a given label."""
        coll = yield self.secretservice.create_collection(label)
        if make_alias:
            coll_path = make_coll_path(label)
            self.mock_service.SetAlias("default", coll_path)
        if publish_default_path:
            mock_object = self.mock_service.collections[label]
            mock_object.add_to_connection(self.session_bus,
                                          txsecrets.DEFAULT_COLLECTION)
            self.addCleanup(mock_object.remove_from_connection,
                            connection=self.session_bus,
                            path=txsecrets.DEFAULT_COLLECTION)
        returnValue(coll)


class SecretServiceTestCase(BaseTestCase):
    """Test the Secret Service class."""

    @inlineCallbacks
    def test_open_session(self):
        """The secret service session is opened."""
        result = yield self.secretservice.open_session()
        self.assertEqual(result, self.secretservice)

    @inlineCallbacks
    def test_open_session_throws_dbus_error_as_failure(self):
        """The secret service open session throws a dbus error as a failure."""
        d = self.secretservice.open_session()
        self.mock_service.open_session_fail = True
        yield self.assertFailure(d, dbus.exceptions.DBusException)

    @inlineCallbacks
    def test_open_session_fails_before_opening_as_failure(self):
        """A dbus error before opening the session is thrown as a failure."""

        def fail(*args, **kwargs):
            """Throw a DBus exception."""
            raise dbus.exceptions.DBusException()

        self.patch(txsecrets.dbus, "SessionBus", fail)
        d = self.secretservice.open_session()
        self.mock_service.open_session_fail = True
        yield self.assertFailure(d, dbus.exceptions.DBusException)

    @inlineCallbacks
    def test_create_collection(self):
        """The secret service creates a collection."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        self.assertIn(collection_label, self.mock_service.collections)

    @inlineCallbacks
    def test_create_collection_prompt(self):
        """The secret service creates a collection after a prompt."""
        yield self.secretservice.open_session()
        self.mock_service.create_collection_prompt = True
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        self.assertIn(collection_label, self.mock_service.collections)

    @inlineCallbacks
    def test_create_collection_prompt_dismissed(self):
        """The service fails to create collection when prompt dismissed."""
        yield self.secretservice.open_session()
        self.mock_service.create_collection_prompt = True
        self.mock_service.dismissed = True
        collection_label = "sample_keyring"
        yield self.assertFailure(
            self.create_sample_collection(collection_label),
            txsecrets.UserCancelled)

    @inlineCallbacks
    def test_create_collection_throws_dbus_error(self):
        """The service fails to create collection on a DBus error."""
        yield self.secretservice.open_session()
        self.mock_service.create_collection_fail = True
        collection_label = "sample_keyring"
        yield self.assertFailure(
            self.create_sample_collection(collection_label),
            dbus.exceptions.DBusException)

    @inlineCallbacks
    def test_prompt_accepted(self):
        """A prompt is accepted."""
        yield self.secretservice.open_session()
        expected_result = "hello world"
        prompt_path = "/prompt"
        self.dbus_publish(prompt_path, PromptMock, result=expected_result,
                          dismissed=False)
        result = yield self.secretservice.do_prompt(prompt_path)
        self.assertEqual(result, expected_result)

    @inlineCallbacks
    def test_prompt_dismissed(self):
        """A prompt is dismissed with a UserCancelled failure."""
        yield self.secretservice.open_session()
        expected_result = "hello world2"
        prompt_path = "/prompt"
        self.dbus_publish(prompt_path, PromptMock, result=expected_result,
                          dismissed=True)
        d = self.secretservice.do_prompt(prompt_path)
        self.assertFailure(d, txsecrets.UserCancelled)

    @inlineCallbacks
    def test_search_unlocked_items(self):
        """The secret service searchs for unlocked items."""
        yield self.secretservice.open_session()
        coll = yield self.create_sample_collection("sample_keyring")
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        value = yield items[0].get_value()
        self.assertEqual(value, sample_secret)

    @inlineCallbacks
    def test_search_locked_items(self):
        """The secret service searchs for locked items."""
        yield self.secretservice.open_session()
        collection_name = "sample_keyring"
        coll = yield self.create_sample_collection(collection_name)
        mock_collection = self.mock_service.collections[collection_name]
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret99!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        mock_collection.locked = True

        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        value = yield items[0].get_value()
        self.assertEqual(value, sample_secret)

    @inlineCallbacks
    def test_search_locked_items_prompts(self):
        """The secret service searchs for locked items after a prompt."""
        yield self.secretservice.open_session()
        collection_name = "sample_keyring"
        coll = yield self.create_sample_collection(collection_name)
        mock_collection = self.mock_service.collections[collection_name]
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret99!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        mock_collection.locked = True
        mock_collection.unlock_prompts = True

        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        value = yield items[0].get_value()
        self.assertEqual(value, sample_secret)

    @inlineCallbacks
    def test_search_locked_items_prompts_dismissed(self):
        """Service fails search for locked items after dismissed prompt."""
        yield self.secretservice.open_session()
        collection_name = "sample_keyring"
        coll = yield self.create_sample_collection(collection_name)
        mock_collection = self.mock_service.collections[collection_name]
        self.mock_service.dismissed = True
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret99!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        mock_collection.locked = True
        mock_collection.unlock_prompts = True

        d = self.secretservice.search_items(attr)
        yield self.assertFailure(d, txsecrets.UserCancelled)

    @inlineCallbacks
    def test_search_items_merges_unlocked_and_locked_items(self):
        """search_items merges unlocked and locked items."""
        yield self.secretservice.open_session()
        attr = {"key-type": "Ubuntu SSO credentials"}

        collection_name = "coll1"
        coll = yield self.create_sample_collection(collection_name)
        mock_coll1 = self.mock_service.collections[collection_name]
        unlocked_secret = "coll 1 secret!"
        yield coll.create_item("Cucaracha", attr, unlocked_secret)
        mock_coll1.locked = False
        mock_coll1.unlock_prompts = False

        collection_name = "coll2"
        coll = yield self.create_sample_collection(collection_name)
        mock_coll2 = self.mock_service.collections[collection_name]
        locked_secret = "coll 2 secret!"
        yield coll.create_item("Cucaracha", attr, locked_secret)
        mock_coll2.locked = True
        mock_coll2.unlock_prompts = False

        result = yield self.secretservice.search_items(attr)
        self.assertEqual(len(result), 2)

    @inlineCallbacks
    def test_search_items_merges_unlocked_locked_and_prompt_items(self):
        """search_items merges unlocked, locked and prompt items."""
        yield self.secretservice.open_session()
        attr = {"key-type": "Ubuntu SSO credentials"}

        collection_name = "coll1"
        coll = yield self.create_sample_collection(collection_name)
        mock_coll1 = self.mock_service.collections[collection_name]
        unlocked_secret = "coll 1 secret!"
        yield coll.create_item("Cucaracha", attr, unlocked_secret)
        mock_coll1.locked = False
        mock_coll1.unlock_prompts = False

        collection_name = "coll2"
        coll = yield self.create_sample_collection(collection_name)
        mock_coll2 = self.mock_service.collections[collection_name]
        locked_secret = "coll 2 secret!"
        yield coll.create_item("Cucaracha", attr, locked_secret)
        mock_coll2.locked = True
        mock_coll2.unlock_prompts = False

        collection_name = "coll3"
        coll = yield self.create_sample_collection(collection_name)
        mock_coll3 = self.mock_service.collections[collection_name]
        locked_secret = "coll 3 secret!"
        yield coll.create_item("Cucaracha", attr, locked_secret)
        mock_coll3.locked = True
        mock_coll3.unlock_prompts = True

        result = yield self.secretservice.search_items(attr)
        self.assertEqual(len(result), 3)

    @inlineCallbacks
    def test_get_collections(self):
        """The list of all collections is returned."""
        collection_names = ["collection1", "collection2"]

        yield self.secretservice.open_session()
        for name in collection_names:
            yield self.create_sample_collection(name)
        collections = yield self.secretservice.get_collections()
        self.assertEqual(len(collections), len(collection_names))

    @inlineCallbacks
    def test_get_default_collection_honours_default_path(self):
        """The default collection is returned from the default path."""
        yield self.secretservice.open_session()
        collection_name = "sample_default_keyring"
        yield self.create_sample_collection(collection_name, make_alias=False,
                                            publish_default_path=True)
        self.assertEqual(len(self.mock_service.collections), 1)
        yield self.secretservice.get_default_collection()
        self.assertEqual(len(self.mock_service.collections), 1)

    @inlineCallbacks
    def test_get_default_collection_honours_readalias(self):
        """The default collection is returned if default alias set."""
        yield self.secretservice.open_session()
        collection_name = "sample_default_keyring"
        yield self.create_sample_collection(collection_name)
        self.assertEqual(len(self.mock_service.collections), 1)
        yield self.secretservice.get_default_collection()
        self.assertEqual(len(self.mock_service.collections), 1)

    @inlineCallbacks
    def test_get_default_collection_created_if_no_default(self):
        """The default collection is created if there's no default."""
        yield self.secretservice.open_session()
        collection_name = "sample_nondefault_keyring"
        yield self.create_sample_collection(collection_name, make_alias=False)
        self.assertEqual(len(self.mock_service.collections), 1)
        yield self.secretservice.get_default_collection()
        self.assertEqual(len(self.mock_service.collections), 2)

    @inlineCallbacks
    def test_get_default_collection_created_if_nonexistent(self):
        """The default collection is created if it doesn't exist yet."""
        yield self.secretservice.open_session()
        self.assertEqual(len(self.mock_service.collections), 0)
        yield self.secretservice.get_default_collection()
        self.assertEqual(len(self.mock_service.collections), 1)

    @inlineCallbacks
    def test_get_default_collection_set_as_default_if_nonexistent(self):
        """The default collection is set as default if it doesn't exist yet."""
        yield self.secretservice.open_session()
        yield self.secretservice.get_default_collection()
        self.assertIn(txsecrets.DEFAULT_LABEL, self.mock_service.aliases)

    @inlineCallbacks
    def test_get_default_collection_is_unlocked_default_path(self):
        """The default collection is unlocked before being returned."""
        yield self.secretservice.open_session()
        collection_name = "sample_keyring"
        self.assertEqual(len(self.mock_service.collections), 0)
        coll = yield self.create_sample_collection(collection_name,
                                                   make_alias=False,
                                                   publish_default_path=True)
        self.assertEqual(len(self.mock_service.collections), 1)
        mock_collection = self.mock_service.collections[collection_name]
        mock_collection.locked = True
        yield self.secretservice.get_default_collection()
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        self.assertEqual(len(mock_collection.items), 1)

    @inlineCallbacks
    def test_get_default_collection_is_unlocked_readalias(self):
        """The default collection is unlocked before being returned."""
        yield self.secretservice.open_session()
        collection_name = "sample_keyring"
        self.assertEqual(len(self.mock_service.collections), 0)
        coll = yield self.create_sample_collection(collection_name)
        self.assertEqual(len(self.mock_service.collections), 1)
        mock_collection = self.mock_service.collections[collection_name]
        mock_collection.locked = True
        yield self.secretservice.get_default_collection()
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        self.assertEqual(len(mock_collection.items), 1)


class CollectionTestCase(BaseTestCase):
    """Test the Collection class."""

    @inlineCallbacks
    def test_get_label(self):
        """The collection gets its own label from the keyring."""
        yield self.secretservice.open_session()
        expected_label = "sample_keyring"
        yield self.create_sample_collection(expected_label)
        coll = yield self.secretservice.get_default_collection()
        result = yield coll.get_label()
        self.assertEqual(result, expected_label)

    @inlineCallbacks
    def test_create_item(self):
        """The collection creates an item."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        coll = yield self.secretservice.get_default_collection()
        mock_collection = self.mock_service.collections[collection_label]
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        self.assertEqual(len(mock_collection.items), 1)
        self.assertEqual(mock_collection.items[0].value, sample_secret)

    @inlineCallbacks
    def test_create_item_prompt(self):
        """The collection creates an item after a prompt."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        coll = yield self.secretservice.get_default_collection()
        mock_collection = self.mock_service.collections[collection_label]
        mock_collection.create_item_prompt = True
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret2!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        self.assertEqual(len(mock_collection.items), 1)
        self.assertEqual(mock_collection.items[0].value, sample_secret)

    @inlineCallbacks
    def test_create_item_prompt_dismissed(self):
        """The collection fails to create an item when prompt is dismissed."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        coll = yield self.secretservice.get_default_collection()
        mock_collection = self.mock_service.collections[collection_label]
        mock_collection.create_item_prompt = True
        mock_collection.dismissed = True
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret3!"
        yield self.assertFailure(coll.create_item("Cuca", attr, sample_secret),
                                 txsecrets.UserCancelled)

    @inlineCallbacks
    def test_create_item_throws_dbus_error(self):
        """The collection fails to create an item when DBus fails."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        coll = yield self.secretservice.get_default_collection()
        mock_collection = self.mock_service.collections[collection_label]
        mock_collection.create_item_fail = True
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret4!"
        yield self.assertFailure(coll.create_item("Cuca", attr, sample_secret),
                                 dbus.exceptions.DBusException)


class ItemTestCase(BaseTestCase):
    """Test the Item class."""

    @inlineCallbacks
    def test_get_value(self):
        """The secret value is retrieved from the item."""
        yield self.secretservice.open_session()
        coll = yield self.create_sample_collection("sample_keyring")
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        value = yield items[0].get_value()
        self.assertEqual(value, sample_secret)

    @inlineCallbacks
    def test_get_value_throws_dbus_error(self):
        """The secret value is not retrieved if DBus fails."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        coll = yield self.create_sample_collection(collection_label)
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        mock = self.mock_service.collections[collection_label].items[0]
        mock.get_secret_fail = True
        yield self.assertFailure(items[0].get_value(),
                                 dbus.exceptions.DBusException)

    @inlineCallbacks
    def test_delete(self):
        """The item is deleted."""
        yield self.secretservice.open_session()
        coll = yield self.create_sample_collection("sample_keyring")
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        yield items[0].delete()
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 0)

    @inlineCallbacks
    def test_delete_prompt(self):
        """The item is deleted after a prompt."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        coll = yield self.create_sample_collection(collection_label)
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        mock_item = self.mock_service.collections[collection_label].items[0]
        mock_item.delete_prompt = True
        yield items[0].delete()
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 0)

    @inlineCallbacks
    def test_delete_prompt_dismissed(self):
        """The item is not deleted after a dismissed prompt."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        coll = yield self.create_sample_collection(collection_label)
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        mock_item = self.mock_service.collections[collection_label].items[0]
        mock_item.delete_prompt = True
        mock_item.dismissed = True
        yield self.assertFailure(items[0].delete(), txsecrets.UserCancelled)

    @inlineCallbacks
    def test_delete_throws_dbus_error(self):
        """The item is not deleted when a DBus error happens."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        coll = yield self.create_sample_collection(collection_label)
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        mock_item = self.mock_service.collections[collection_label].items[0]
        mock_item.delete_fail = True
        yield self.assertFailure(items[0].delete(),
                                 dbus.exceptions.DBusException)


class AltItemTestCase(BaseTestCase):
    """Test the Item class with 4 fields in the secret struct."""
    secret_service_class = AltSecretServiceMock

    @inlineCallbacks
    def test_create_item_four_fields_per_secret(self):
        """The collection creates an item when the dbus struct has 4 fields."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        yield self.create_sample_collection(collection_label)
        coll = yield self.secretservice.get_default_collection()
        mock_collection = self.mock_service.collections[collection_label]
        attr = {"key-type": "Ubuntu 242 credentials"}
        sample_secret = "secret!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        self.assertEqual(len(mock_collection.items), 1)
        self.assertEqual(mock_collection.items[0].value, sample_secret)

    @inlineCallbacks
    def test_get_value_four_fields_per_secret(self):
        """The code works fine when the secret dbus struct has 4 fields."""
        yield self.secretservice.open_session()
        collection_label = "sample_keyring"
        coll = yield self.create_sample_collection(collection_label)
        attr = {"key-type": "Ubuntu SSO credentials"}
        sample_secret = "secret83!"
        yield coll.create_item("Cucaracha", attr, sample_secret)
        items = yield self.secretservice.search_items(attr)
        self.assertEqual(len(items), 1)
        value = yield items[0].get_value()
        self.assertEqual(value, sample_secret)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the Setup Account page Qt UI."""

from PyQt4 import QtCore

from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.qt import forgotten_password_page as gui
from ubuntu_kylin_sso.qt.tests import (
    PageBaseTestCase,
    FakePageUiStyle,
)

# is ok to access private method/attrs in tests
# pylint: disable=W0212, E1101


class ForgottenPasswordTestCase(PageBaseTestCase):
    """Test the SetupAccountPage code."""

    ui_class = gui.ForgottenPasswordPage
    ui_siganls = ('passwordResetTokenSent',)
    ui_backend_siganls = ('PasswordResetTokenSent', 'PasswordResetError')

    def test_request_new_password(self):
        """Test the request_new_password function."""
        email = 'email@example.com'
        self.ui.ui.email_line_edit.setText(email)
        self.ui.request_new_password()
        self.assertEqual(self._overlay_show_counter, 1)
        self.assertFalse(self.ui.isEnabled())
        self.assert_backend_called('request_password_reset_token',
            self.app_name, email)

    def test_email_address(self):
        """Test the email_address property."""
        value = self.ui.email_address
        self.assertEqual(value,
                         compat.text_type(self.ui.ui.email_line_edit.text()))
        self.assertTrue(isinstance(value, compat.text_type))

    def test_send_button_clicked(self):
        """Test the send_button property."""
        self.patch(self.ui, "hide_error", self._set_called)
        self.ui.ui.send_button.clicked.emit(True)
        self.assert_backend_called('request_password_reset_token',
            self.app_name, '')
        expected = ((), {})
        self.assertEqual(expected, self._called)

    def test_initialize_page_with_email(self):
        """Test the initialization method."""
        self.ui.ui.email_line_edit.setText('mail@example')
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "send_button", button)
        self.ui.initializePage()
        self.assertTrue(button.properties['default'])
        self.assertTrue(button.isEnabled())

    def test_set_translated_strings(self):
        """Test the translated string method."""
        self.assert_title_correct(gui.FORGOTTEN_PASSWORD_TITLE)
        subtitle = gui.FORGOTTEN_PASSWORD_SUBTITLE
        self.assert_subtitle_correct(subtitle.format(app_name=self.app_name))

        self.assertEqual(
                compat.text_type(self.ui.ui.email_address_label.text()),
                gui.EMAIL_LABEL)
        self.assertEqual(compat.text_type(self.ui.ui.send_button.text()),
            gui.RESET_PASSWORD)

    def test_connect_ui(self):
        """Test the connect ui method."""
        self.assertEqual(self.ui.ui.email_line_edit.receivers(
            QtCore.SIGNAL("textChanged(const QString&)")), 2)
        self.assertEqual(self.ui.ui.send_button.receivers(
            QtCore.SIGNAL("clicked()")), 1)

    def test_validate_not_valid(self):
        """Test validate method."""
        self.ui.ui.email_line_edit.setText('')
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "send_button", button)
        self.ui._validate()
        self.assertFalse(button.isEnabled())

    def test_validate_valid(self):
        """Test validate method."""
        self.ui.ui.email_line_edit.setText('mail@example')
        button = FakePageUiStyle()
        self.patch(self.ui.ui, "send_button", button)
        self.ui._validate()
        self.assertTrue(button.isEnabled())

    def test_on_password_reset_token_sent(self):
        """Test on_password_reset_token_sent method."""
        email = 'foo@ubuntu.com'
        self.assert_signal_emitted(self.ui.passwordResetTokenSent, (email,),
            self.ui.on_password_reset_token_sent, self.app_name, email)

    def test_on_password_reset_error(self):
        """Test on_password_reset_error method."""
        self.patch(self.ui, "show_error", self._set_called)
        error = {'errtype': 'FooBarBaz'}

        self.ui.on_password_reset_error(self.app_name, error)

        expected = ((self.ui, gui.REQUEST_PASSWORD_TOKEN_WRONG_EMAIL), {})
        self.assertTrue(expected, self._called)
